import dataclasses
import os
import stat
import subprocess
import sys
import textwrap
from typing import Optional, List, Union, NoReturn, Mapping

from debputy.commands.debputy_cmd.context import CommandContext
from debputy.commands.debputy_cmd.output import _output_styling, OutputStylingBase
from debputy.filesystem_scan import FSROOverlay
from debputy.linting.lint_util import (
    LintReport,
    LintStateImpl,
    FormatterImpl,
    TermLintReport,
    LintDiagnosticResultState,
    AsyncLinterImpl,
)
from debputy.lsp.diagnostics import DiagnosticData
from debputy.lsp.lsp_debian_changelog import _lint_debian_changelog
from debputy.lsp.lsp_debian_control import (
    _lint_debian_control,
    _reformat_debian_control,
)
from debputy.lsp.lsp_debian_copyright import (
    _lint_debian_copyright,
    _reformat_debian_copyright,
)
from debputy.lsp.lsp_debian_debputy_manifest import _lint_debian_debputy_manifest
from debputy.lsp.lsp_debian_patches_series import _lint_debian_patches_series
from debputy.lsp.lsp_debian_rules import _lint_debian_rules
from debputy.lsp.lsp_debian_tests_control import (
    _lint_debian_tests_control,
    _reformat_debian_tests_control,
)
from debputy.lsp.lsp_debian_upstream_metadata import _lint_debian_upstream_metadata
from debputy.lsp.maint_prefs import (
    MaintainerPreferenceTable,
    EffectiveFormattingPreference,
    determine_effective_preference,
)
from debputy.lsp.quickfixes import provide_standard_quickfixes_from_diagnostics_lint
from debputy.lsp.spellchecking import disable_spellchecking
from debputy.lsp.text_edit import (
    get_well_formatted_edit,
    merge_sort_text_edits,
    apply_text_edits,
    OverLappingTextEditException,
)
from debputy.lsp.vendoring._deb822_repro import Deb822FileElement
from debputy.lsprotocol.types import (
    CodeAction,
    Command,
    CodeActionParams,
    CodeActionContext,
    TextDocumentIdentifier,
    TextEdit,
    Position,
    Range,
    DiagnosticSeverity,
    Diagnostic,
)
from debputy.packages import SourcePackage, BinaryPackage
from debputy.plugin.api import VirtualPath
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from debputy.util import _warn, _error, _info
from debputy.yaml import MANIFEST_YAML, YAMLError
from debputy.yaml.compat import CommentedMap

LINTER_FORMATS = {
    "debian/changelog": _lint_debian_changelog,
    "debian/control": _lint_debian_control,
    "debian/copyright": _lint_debian_copyright,
    "debian/debputy.manifest": _lint_debian_debputy_manifest,
    "debian/upstream/metadata": _lint_debian_upstream_metadata,
    "debian/rules": _lint_debian_rules,
    "debian/patches/series": _lint_debian_patches_series,
    "debian/tests/control": _lint_debian_tests_control,
}


REFORMAT_FORMATS = {
    "debian/control": _reformat_debian_control,
    "debian/copyright": _reformat_debian_copyright,
    "debian/tests/control": _reformat_debian_tests_control,
}


@dataclasses.dataclass(slots=True)
class LintContext:
    plugin_feature_set: PluginProvidedFeatureSet
    maint_preference_table: MaintainerPreferenceTable
    source_root: Optional[VirtualPath]
    debian_dir: Optional[VirtualPath]
    parsed_deb822_file_content: Optional[Deb822FileElement] = None
    source_package: Optional[SourcePackage] = None
    binary_packages: Optional[Mapping[str, BinaryPackage]] = None
    effective_preference: Optional[EffectiveFormattingPreference] = None
    style_tool: Optional[str] = None
    unsupported_preference_reason: Optional[str] = None
    salsa_ci: Optional[CommentedMap] = None

    def state_for(
        self,
        path: str,
        content: str,
        lines: List[str],
        lint_implementation: Optional[AsyncLinterImpl],
    ) -> LintStateImpl:
        return LintStateImpl(
            self.plugin_feature_set,
            self.maint_preference_table,
            self.source_root,
            self.debian_dir,
            path,
            content,
            lines,
            self.source_package,
            self.binary_packages,
            self.effective_preference,
            lint_implementation,
        )


def gather_lint_info(context: CommandContext) -> LintContext:
    source_root = FSROOverlay.create_root_dir(".", ".")
    debian_dir = source_root.get("debian")
    if debian_dir is not None and not debian_dir.is_dir:
        debian_dir = None
    lint_context = LintContext(
        context.load_plugins(),
        MaintainerPreferenceTable.load_preferences(),
        source_root,
        debian_dir,
    )
    try:
        with open("debian/control") as fd:
            deb822_file, source_package, binary_packages = (
                context.dctrl_parser.parse_source_debian_control(fd, ignore_errors=True)
            )
    except FileNotFoundError:
        source_package = None
    else:
        lint_context.parsed_deb822_file_content = deb822_file
        lint_context.source_package = source_package
        lint_context.binary_packages = binary_packages
    salsa_ci_map: Optional[CommentedMap] = None
    for ci_file in ("debian/salsa-ci.yml", ".gitlab-ci.yml"):
        try:
            with open(ci_file) as fd:
                salsa_ci_map = MANIFEST_YAML.load(fd)
                if not isinstance(salsa_ci_map, CommentedMap):
                    salsa_ci_map = None
                break
        except FileNotFoundError:
            pass
        except YAMLError:
            break
    if source_package is not None or salsa_ci_map is not None:
        pref, tool, pref_reason = determine_effective_preference(
            lint_context.maint_preference_table,
            source_package,
            salsa_ci_map,
        )
        lint_context.effective_preference = pref
        lint_context.style_tool = tool
        lint_context.unsupported_preference_reason = pref_reason

    return lint_context


def initialize_lint_report(context: CommandContext) -> LintReport:
    lint_report_format = context.parsed_args.lint_report_format
    report_output = context.parsed_args.report_output

    if lint_report_format == "term":
        fo = _output_styling(context.parsed_args, sys.stdout)
        if report_output is not None:
            _warn("--report-output is redundant for the `term` report")
        return TermLintReport(fo)
    if lint_report_format == "junit4-xml":
        try:
            import junit_xml
        except ImportError:
            _error(
                "The `junit4-xml` report format requires `python3-junit.xml` to be installed"
            )

        from debputy.linting.lint_report_junit import JunitLintReport

        if report_output is None:
            report_output = "debputy-lint-junit.xml"

        return JunitLintReport(report_output)

    raise AssertionError(f"Missing case for lint_report_format: {lint_report_format}")


async def perform_linting(context: CommandContext) -> None:
    parsed_args = context.parsed_args
    if not parsed_args.spellcheck:
        disable_spellchecking()
    linter_exit_code = parsed_args.linter_exit_code
    lint_report = initialize_lint_report(context)
    lint_context = gather_lint_info(context)

    for name_stem in LINTER_FORMATS:
        filename = f"./{name_stem}"
        if not os.path.isfile(filename):
            continue
        await perform_linting_of_file(
            lint_context,
            filename,
            name_stem,
            context.parsed_args.auto_fix,
            lint_report,
        )
    if lint_report.number_of_invalid_diagnostics:
        _warn(
            "Some diagnostics did not explicitly set severity. Please report the bug and include the output"
        )
    if lint_report.number_of_broken_diagnostics:
        _error(
            "Some sub-linters reported issues. Please report the bug and include the output"
        )

    if parsed_args.warn_about_check_manifest and os.path.isfile(
        "debian/debputy.manifest"
    ):
        _info("Note: Due to a limitation in the linter, debian/debputy.manifest is")
        _info("only **partially** checked by this command at the time of writing.")
        _info("Please use `debputy check-manifest` to fully check the manifest.")

    lint_report.finish_report()

    if linter_exit_code:
        _exit_with_lint_code(lint_report)


def perform_reformat(
    context: CommandContext,
    *,
    named_style: Optional[str] = None,
) -> None:
    parsed_args = context.parsed_args
    fo = _output_styling(context.parsed_args, sys.stdout)
    lint_context = gather_lint_info(context)
    if named_style is not None:
        style = lint_context.maint_preference_table.named_styles.get(named_style)
        if style is None:
            styles = ", ".join(lint_context.maint_preference_table.named_styles)
            _error(f'There is no style named "{style}". Options include: {styles}')
        if (
            lint_context.effective_preference is not None
            and lint_context.effective_preference != style
        ):
            _info(
                f'Note that the style "{named_style}" does not match the style that `debputy` was configured to use.'
            )
            _info("This may be a non-issue (if the configuration is out of date).")
        lint_context.effective_preference = style

    if lint_context.effective_preference is None:
        if lint_context.unsupported_preference_reason is not None:
            _warn(
                "While `debputy` could identify a formatting for this package, it does not support it."
            )
            _warn(f"{lint_context.unsupported_preference_reason}")
            if lint_context.style_tool is not None:
                _info(
                    f"The following tool might be able to apply the style: {lint_context.style_tool}"
                )
            if parsed_args.supported_style_required:
                _error(
                    "Sorry; `debputy` does not support the style. Use --unknown-or-unsupported-style-is-ok to make"
                    " this a non-error (note that `debputy` will not reformat the packaging in this case; just not"
                    " exit with an error code)."
                )
        else:
            print(
                textwrap.dedent(
                    """\
                You can enable set a style by doing either of:

                 * You can set `X-Style: black` in the source stanza of `debian/control` to pick
                   `black` as the preferred style for this package.
                   - Note: `black` is an opinionated style that follows the spirit of the `black` code formatter
                     for Python.
                   - If you use `pre-commit`, then there is a formatting hook at
                     https://salsa.debian.org/debian/debputy-pre-commit-hooks

                 * If you use the Debian Salsa CI pipeline, then you can set SALSA_CI_DISABLE_WRAP_AND_SORT (`no`)
                   or SALSA_CI_ENABLE_WRAP_AND_SORT (`yes`) to the relevant value and `debputy` will pick up the
                   configuration from there.
                   - Note: The option must be in `.gitlab-ci.yml` or `debian/salsa-ci.yml` to work. The Salsa CI
                     pipeline will use `wrap-and-sort` while `debputy` uses its own emulation of `wrap-and-sort`
                     (`debputy` also needs to apply the style via `debputy lsp server`).

                 * The `debputy` code also comes with a built-in style database. This may be interesting for
                   packaging teams, so set a default team style that applies to all packages maintained by
                   that packaging team.
                   - Individuals can also add their style, which can useful for ad-hoc packaging teams, where
                     `debputy` will automatically apply a style if *all* co-maintainers agree to it.

                Note the above list is an ordered list of how `debputy` determines which style to use in case
                multiple options are available.
                """
                )
            )
            if parsed_args.supported_style_required:
                if lint_context.style_tool is not None:
                    _error(
                        "Sorry, `debputy reformat` does not support the packaging style. However, the"
                        f" formatting is supposedly handled by: {lint_context.style_tool}"
                    )
                _error(
                    "Sorry; `debputy` does not know which style to use for this package. Please either set a"
                    "style or use --unknown-or-unsupported-style-is-ok to make this a non-error"
                )
        _info("")
        _info(
            "Doing nothing since no supported style could be identified as requested."
            " See above how to set a style."
        )
        _info("Use --supported-style-is-required if this should be an error instead.")
        sys.exit(0)

    changes = False
    auto_fix = context.parsed_args.auto_fix
    for name_stem in REFORMAT_FORMATS:
        formatter = REFORMAT_FORMATS.get(name_stem)
        filename = f"./{name_stem}"
        if formatter is None or not os.path.isfile(filename):
            continue

        reformatted = perform_reformat_of_file(
            fo,
            lint_context,
            filename,
            formatter,
            auto_fix,
        )
        if reformatted:
            changes = True

    if changes and parsed_args.linter_exit_code:
        sys.exit(2)


def perform_reformat_of_file(
    fo: OutputStylingBase,
    lint_context: LintContext,
    filename: str,
    formatter: FormatterImpl,
    auto_fix: bool,
) -> bool:
    with open(filename, "rt", encoding="utf-8") as fd:
        text = fd.read()

    lines = text.splitlines(keepends=True)
    lint_state = lint_context.state_for(
        filename,
        text,
        lines,
        None,
    )
    edits = formatter(lint_state)
    if not edits:
        return False

    try:
        replacement = apply_text_edits(text, lines, edits)
    except OverLappingTextEditException:
        _error(
            f"The reformater for {filename} produced overlapping edits (which is broken and will not work)"
        )

    output_filename = f"{filename}.tmp"
    with open(output_filename, "wt", encoding="utf-8") as fd:
        fd.write(replacement)

    r = subprocess.run(["diff", "-u", filename, output_filename]).returncode
    if r != 0 and r != 1:
        _warn(f"diff -u {filename} {output_filename} failed!?")
    if auto_fix:
        orig_mode = stat.S_IMODE(os.stat(filename).st_mode)
        os.chmod(output_filename, orig_mode)
        os.rename(output_filename, filename)
        print(
            fo.colored(
                f"Reformatted {filename}.",
                fg="green",
                style="bold",
            )
        )
    else:
        os.unlink(output_filename)

    return True


def _exit_with_lint_code(lint_report: LintReport) -> NoReturn:
    diagnostics_count = lint_report.diagnostics_count
    if (
        diagnostics_count[DiagnosticSeverity.Error]
        or diagnostics_count[DiagnosticSeverity.Warning]
    ):
        sys.exit(2)
    sys.exit(0)


async def perform_linting_of_file(
    lint_context: LintContext,
    filename: str,
    file_format: str,
    auto_fixing_enabled: bool,
    lint_report: LintReport,
) -> None:
    handler = LINTER_FORMATS.get(file_format)
    if handler is None:
        return
    with open(filename, "rt", encoding="utf-8") as fd:
        text = fd.read()

    if auto_fixing_enabled:
        await _auto_fix_run(
            lint_context,
            filename,
            text,
            handler,
            lint_report,
        )
    else:
        await _diagnostics_run(
            lint_context,
            filename,
            text,
            handler,
            lint_report,
        )


def _overlapping_edit(
    last_edit_range: Range,
    last_fix_range: Range,
) -> bool:
    last_edit_start_pos = last_edit_range.start
    last_fix_start_pos = last_fix_range.start
    if last_edit_start_pos.line < last_fix_start_pos.line:
        return True
    if (
        last_edit_start_pos.line == last_fix_start_pos.character
        and last_edit_start_pos.character < last_fix_start_pos.character
    ):
        return True

    if last_fix_range.end == last_fix_start_pos:
        return False

    if last_fix_range.end < last_edit_start_pos:
        return False
    return True


def _max_range(
    r1: Range,
    r2: Range,
) -> Range:
    if r2.end > r1.end:
        return r2
    if r2.end < r1.end:
        return r1
    if r2.start > r1.start:
        return r2
    return r1


_INITIAL_FIX_RANGE = Range(
    Position(0, 0),
    Position(0, 0),
)


def _is_non_interactive_auto_fix_allowed(diagnostic: Diagnostic) -> bool:
    diag_data: Optional[DiagnosticData] = diagnostic.data
    return diag_data is None or diag_data.get("enable_non_interactive_auto_fix", True)


async def _auto_fix_run(
    lint_context: LintContext,
    filename: str,
    text: str,
    linter: AsyncLinterImpl,
    lint_report: LintReport,
) -> None:
    another_round = True
    unfixed_diagnostics: List[Diagnostic] = []
    remaining_rounds = 10
    fixed_count = 0
    too_many_rounds = False
    lines = text.splitlines(keepends=True)
    lint_state = lint_context.state_for(
        filename,
        text,
        lines,
        linter,
    )
    current_issues = await lint_state.gather_diagnostics()
    issue_count_start = len(current_issues) if current_issues else 0
    while another_round and current_issues:
        another_round = False
        last_fix_range = _INITIAL_FIX_RANGE
        unfixed_diagnostics.clear()
        edits = []
        fixed_diagnostics = []
        for diagnostic in current_issues:
            if not _is_non_interactive_auto_fix_allowed(diagnostic):
                unfixed_diagnostics.append(diagnostic)
                continue
            actions = provide_standard_quickfixes_from_diagnostics_lint(
                CodeActionParams(
                    TextDocumentIdentifier(filename),
                    diagnostic.range,
                    CodeActionContext(
                        [diagnostic],
                    ),
                ),
            )
            auto_fixing_edits = resolve_auto_fixer(filename, actions)

            if not auto_fixing_edits:
                unfixed_diagnostics.append(diagnostic)
                continue

            sorted_edits = merge_sort_text_edits(
                [get_well_formatted_edit(e) for e in auto_fixing_edits],
            )
            last_edit = sorted_edits[-1]
            last_edit_range = last_edit.range
            if _overlapping_edit(last_edit_range, last_fix_range):
                if not another_round:
                    if remaining_rounds > 0:
                        remaining_rounds -= 1
                        print(
                            "Detected overlapping edit; scheduling another edit round."
                        )
                        another_round = True
                    else:
                        _warn(
                            "Too many overlapping edits; stopping after this round (circuit breaker)."
                        )
                        too_many_rounds = True
                continue
            edits.extend(sorted_edits)
            fixed_diagnostics.append(diagnostic)
            last_fix_range = _max_range(last_fix_range, sorted_edits[-1].range)

        if another_round and not edits:
            _error(
                "Internal error: Detected an overlapping edit and yet had no edits to perform..."
            )

        fixed_count += len(fixed_diagnostics)

        try:
            text = apply_text_edits(
                text,
                lines,
                edits,
            )
        except OverLappingTextEditException:
            _error(
                f"Failed to apply edits for f{filename} due to over lapping edits. Please file a bug"
                f" against `debputy` with a contents of the `debian` directory or a minimal reproducer"
            )
        lines = text.splitlines(keepends=True)

        with lint_report.line_state(lint_state):
            for diagnostic in fixed_diagnostics:
                lint_report.report_diagnostic(
                    diagnostic,
                    result_state=LintDiagnosticResultState.FIXED,
                )
        lint_state.content = text
        lint_state.lines = lines
        lint_state.clear_cache()
        current_issues = await lint_state.gather_diagnostics()

    if fixed_count:
        output_filename = f"{filename}.tmp"
        with open(output_filename, "wt", encoding="utf-8") as fd:
            fd.write(text)
        orig_mode = stat.S_IMODE(os.stat(filename).st_mode)
        os.chmod(output_filename, orig_mode)
        os.rename(output_filename, filename)
        lines = text.splitlines(keepends=True)
        lint_state.content = text
        lint_state.lines = lines
        remaining_issues = await lint_state.gather_diagnostics() or []
    else:
        remaining_issues = current_issues or []

    with lint_report.line_state(lint_state):
        for diagnostic in remaining_issues:
            lint_report.report_diagnostic(diagnostic)

    if isinstance(lint_report, TermLintReport):
        # TODO: Not optimal, but will do for now.
        fo = lint_report.fo
        print()
        if fixed_count:
            remaining_issues_count = len(remaining_issues)
            print(
                fo.colored(
                    f"Fixes applied to {filename}: {fixed_count}."
                    f" Number of issues went from {issue_count_start} to {remaining_issues_count}",
                    fg="green",
                    style="bold",
                )
            )
        elif remaining_issues:
            print(
                fo.colored(
                    f"None of the issues in {filename} could be fixed automatically. Sorry!",
                    fg="yellow",
                    bg="black",
                    style="bold",
                )
            )
        else:
            assert not current_issues
            print(
                fo.colored(
                    f"No issues detected in {filename}",
                    fg="green",
                    style="bold",
                )
            )
        if too_many_rounds:
            print(
                fo.colored(
                    f"Not all fixes for issues in {filename} could be applied due to overlapping edits.",
                    fg="yellow",
                    bg="black",
                    style="bold",
                )
            )
            print(
                "Running once more may cause more fixes to be applied. However, you may be facing"
                " pathological performance."
            )


async def _diagnostics_run(
    lint_context: LintContext,
    filename: str,
    text: str,
    linter: AsyncLinterImpl,
    lint_report: LintReport,
) -> None:
    lines = text.splitlines(keepends=True)
    lint_state = lint_context.state_for(filename, text, lines, linter)
    with lint_report.line_state(lint_state):
        issues = await lint_state.gather_diagnostics()
        for diagnostic in issues:
            actions = provide_standard_quickfixes_from_diagnostics_lint(
                CodeActionParams(
                    TextDocumentIdentifier(filename),
                    diagnostic.range,
                    CodeActionContext(
                        [diagnostic],
                    ),
                ),
            )
            auto_fixer = resolve_auto_fixer(filename, actions)
            has_auto_fixer = bool(auto_fixer) and _is_non_interactive_auto_fix_allowed(
                diagnostic
            )

            result_state = LintDiagnosticResultState.REPORTED
            if has_auto_fixer:
                result_state = LintDiagnosticResultState.AUTO_FIXABLE
            elif has_at_least_lsp_quickfix(actions):
                result_state = LintDiagnosticResultState.MANUAL_FIXABLE

            lint_report.report_diagnostic(diagnostic, result_state=result_state)


def has_at_least_lsp_quickfix(
    actions: Optional[List[Union[Command, CodeAction]]],
) -> bool:
    if actions is None:
        return False
    for action in actions:
        if not isinstance(action, CodeAction):
            continue
        if action.edit is None and action.command is None:
            continue
        return True
    return False


def resolve_auto_fixer(
    document_ref: str,
    actions: Optional[List[Union[Command, CodeAction]]],
) -> Optional[List[TextEdit]]:
    if actions is None or len(actions) != 1:
        return None
    action = actions[0]
    if not isinstance(action, CodeAction):
        return None
    workspace_edit = action.edit
    if workspace_edit is None or action.command is not None:
        return None
    if (
        not workspace_edit.changes
        or len(workspace_edit.changes) != 1
        or document_ref not in workspace_edit.changes
    ):
        return None
    return workspace_edit.changes[document_ref]



import argparse
import json
from pathlib import Path

from debputy.commands import deb_packer
from debputy.intermediate_manifest import TarMember, PathType


def write_unpacked_deb(root: Path, package: str, version: str, arch: str):
    debian = root / "DEBIAN"
    debian.mkdir(mode=0o755)
    (debian / "control").write_text(
        f"Package: {package}\nVersion: {version}\nArchitecture: {arch}\n"
    )


def test_pack_smoke(tmp_path):
    mtime = 1668973695

    root_dir = tmp_path / "root"
    root_dir.mkdir()
    write_unpacked_deb(root_dir, "fake", "1.0", "amd64")
    output_path = tmp_path / "out"
    output_path.mkdir()
    deb_file = Path(output_path) / "output.deb"

    parsed_args = argparse.Namespace(
        is_udeb=False, compression_level=None, compression_strategy=None
    )

    data_compression = deb_packer.COMPRESSIONS["xz"]
    data_compression_cmd = data_compression.as_cmdline(parsed_args)
    ctrl_compression = data_compression
    ctrl_compression_cmd = data_compression_cmd

    package_manifest = tmp_path / "temporary-manifest.json"
    package_manifest.write_text(
        json.dumps(
            [
                TarMember.virtual_path(
                    "./", PathType.DIRECTORY, mode=0o755, mtime=1668973695
                ).to_manifest()
            ]
        )
    )

    deb_packer.pack(
        str(deb_file),
        ctrl_compression,
        data_compression,
        str(root_dir),
        str(package_manifest),
        mtime,
        ctrl_compression_cmd,
        data_compression_cmd,
        prefer_raw_exceptions=True,
    )

    binary = deb_file.read_bytes()

    assert binary == (
        b"!<arch>\n"
        b"debian-binary   1668973695  0     0     100644  4         `\n"
        b"2.0\n"
        b"control.tar.xz  1668973695  0     0     100644  244       `\n"
        b"\xfd7zXZ\x00\x00\x04\xe6\xd6\xb4F\x04\xc0\xb4\x01\x80P!\x01\x16\x00\x00\x00"
        b"\x00\x00\x00\x00\x19\x87 E\xe0'\xff\x00\xac]\x00\x17\x0b\xbc\x1c}"
        b"\x01\x95\xc0\x1dJ>y\x15\xc2\xcc&\xa3^\x11\xb5\x81\xa6\x8cI\xd2\xf0m\xdd\x04"
        b"M\xb2|Tdy\xf5\x00H\xab\xa6B\x11\x8d2\x0e\x1d\xf8F\x9e\x9a\xb0\xb8_]\xa3;M"
        b"t\x90\x9a\xe3)\xeb\xadF\xfet'b\x05\x85\xd5\x04g\x7f\x89\xeb=(\xfd\xf6"
        b'"p\xc3\x91\xf2\xd3\xd2\xb3\xed%i\x9a\xfa\\\xde7\xd5\x01\x18I\x14D\x10E'
        b"\xba\xdf\xfb\x12{\x84\xc4\x10\x08,\xbc\x9e\xac+w\x07\r`|\xcfFL#\xbb"
        b"S\x91\xb4\\\x9b\x80&\x1d\x9ej\x13\xe3\x13\x02=\xe9\xd5\xcf\xb0\xdf?L\xf1\x96"
        b"\xd2\xc6bh\x19|?\xc2j\xe58If\xb7Y\xb9\x18:\x00\x00|\xfb\xcf\x82e/\xd05"
        b"\x00\x01\xd0\x01\x80P\x00\x00\xc9y\xeem\xb1\xc4g\xfb\x02\x00\x00\x00"
        b"\x00\x04YZ"
        b"data.tar.xz     1668973695  0     0     100644  168       `\n"
        b"\xfd7zXZ\x00\x00\x04\xe6\xd6\xb4F\x04\xc0h\x80P!\x01\x16\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\xc2bi\xe8\xe0'\xff\x00`]\x00\x17\x0b\xbc\x1c}"
        b"\x01\x95\xc0\x1dJ>y\x15\xc2\xcc&\xa3^\x11\xb5\x81\xa6\x8cI\xd2\xf0m\xdd\x04"
        b"M\xb2|Tdy\xf5\x00H\xab\xa6B\x11\x8d2\x0e\x1d\xf8F\x9e\x9a\xb0\xb8_]\xa4W%"
        b"\xa2\x14N\xb9\xe7\xbd\xf3a\x16\xe5\xb7\xe6\x80\x95\xcc\xe6+\xe1;I"
        b"\xf2\x1f\xed\x08\xac\xd7UZ\xc0P\x0b\xfb\nK\xef~\xcb\x8f\x80\x00\x9b\x19\xf8A"
        b"Q_\xe7\xeb\x00\x01\x84\x01\x80P\x00\x00(3\xf1\xfa\xb1\xc4g\xfb"
        b"\x02\x00\x00\x00\x00\x04YZ"
    )



import collections
import dataclasses
import os
import textwrap
from typing import Dict, List, Literal, Iterable, Sequence

from debputy.packages import BinaryPackage
from debputy.plugin.api.spec import (
    ServiceRegistry,
    VirtualPath,
    PackageProcessingContext,
    BinaryCtrlAccessor,
    ServiceDefinition,
)
from debputy.util import _error, assume_not_none

DPKG_ROOT = '"${DPKG_ROOT}"'
EMPTY_DPKG_ROOT_CONDITION = '[ -z "${DPKG_ROOT}" ]'
SERVICE_MANAGER_IS_SYSTEMD_CONDITION = "[ -d /run/systemd/system ]"


@dataclasses.dataclass(slots=True)
class SystemdServiceContext:
    had_install_section: bool


@dataclasses.dataclass(slots=True)
class SystemdUnit:
    path: VirtualPath
    names: List[str]
    type_of_service: str
    service_scope: str
    enable_by_default: bool
    start_by_default: bool
    had_install_section: bool


def detect_systemd_service_files(
    fs_root: VirtualPath,
    service_registry: ServiceRegistry[SystemdServiceContext],
    context: PackageProcessingContext,
) -> None:
    pkg = context.binary_package
    systemd_units = _find_and_analyze_systemd_service_files(pkg, fs_root, "system")
    for unit in systemd_units:
        service_registry.register_service(
            unit.path,
            unit.names,
            type_of_service=unit.type_of_service,
            service_scope=unit.service_scope,
            enable_by_default=unit.enable_by_default,
            start_by_default=unit.start_by_default,
            default_upgrade_rule="restart" if unit.start_by_default else "do-nothing",
            service_context=SystemdServiceContext(
                unit.had_install_section,
            ),
        )


def generate_snippets_for_systemd_units(
    services: Sequence[ServiceDefinition[SystemdServiceContext]],
    ctrl: BinaryCtrlAccessor,
    _context: PackageProcessingContext,
) -> None:
    stop_before_upgrade: List[str] = []
    stop_then_start_scripts = []
    on_purge = []
    start_on_install = []
    action_on_upgrade = collections.defaultdict(list)
    assert services

    for service_def in services:
        if service_def.auto_enable_on_install:
            template = """\
                if deb-systemd-helper debian-installed {UNITFILE}; then
                    # The following line should be removed in trixie or trixie+1
                    deb-systemd-helper unmask {UNITFILE} >/dev/null || true

                    if deb-systemd-helper --quiet was-enabled {UNITFILE}; then
                        # Create new symlinks, if any.
                        deb-systemd-helper enable {UNITFILE} >/dev/null || true
                    fi
                fi

                # Update the statefile to add new symlinks (if any), which need to be cleaned
                # up on purge. Also remove old symlinks.
                deb-systemd-helper update-state {UNITFILE} >/dev/null || true
            """
        else:
            template = """\
                # The following line should be removed in trixie or trixie+1
                deb-systemd-helper unmask {UNITFILE} >/dev/null || true

                # was-enabled defaults to true, so new installations run enable.
                if deb-systemd-helper --quiet was-enabled {UNITFILE}; then
                    # Enables the unit on first installation, creates new
                    # symlinks on upgrades if the unit file has changed.
                    deb-systemd-helper enable {UNITFILE} >/dev/null || true
                else
                    # Update the statefile to add new symlinks (if any), which need to be
                    # cleaned up on purge. Also remove old symlinks.
                    deb-systemd-helper update-state {UNITFILE} >/dev/null || true
                fi
            """
        service_name = service_def.name

        if assume_not_none(service_def.service_context).had_install_section:
            ctrl.maintscript.on_configure(
                template.format(
                    UNITFILE=ctrl.maintscript.escape_shell_words(service_name),
                )
            )
            on_purge.append(service_name)
        elif service_def.auto_enable_on_install:
            _error(
                f'The service "{service_name}" cannot be enabled under "systemd" as'
                f' it has no "[Install]" section. Please correct {service_def.definition_source}'
                f' so that it does not enable the service or does not apply to "systemd"'
            )

        if service_def.auto_start_on_install:
            start_on_install.append(service_name)
        if service_def.on_upgrade == "stop-then-start":
            stop_then_start_scripts.append(service_name)
        elif service_def.on_upgrade in ("restart", "reload"):
            action: str = service_def.on_upgrade
            action_on_upgrade[action].append(service_name)
        elif service_def.on_upgrade != "do-nothing":
            raise AssertionError(
                f"Missing support for on_upgrade rule: {service_def.on_upgrade}"
            )

    if start_on_install or action_on_upgrade:
        lines = [
            "if {EMPTY_DPKG_ROOT_CONDITION} && {SERVICE_MANAGER_IS_SYSTEMD_CONDITION}; then".format(
                EMPTY_DPKG_ROOT_CONDITION=EMPTY_DPKG_ROOT_CONDITION,
                SERVICE_MANAGER_IS_SYSTEMD_CONDITION=SERVICE_MANAGER_IS_SYSTEMD_CONDITION,
            ),
            "    systemctl --system daemon-reload >/dev/null || true",
        ]
        if stop_then_start_scripts:
            unit_files = ctrl.maintscript.escape_shell_words(*stop_then_start_scripts)
            lines.append(
                "        deb-systemd-invoke start {UNITFILES} >/dev/null || true".format(
                    UNITFILES=unit_files,
                )
            )
        if start_on_install:
            lines.append('    if [ -z "$2" ]; then')
            lines.append(
                "        deb-systemd-invoke start {UNITFILES} >/dev/null || true".format(
                    UNITFILES=ctrl.maintscript.escape_shell_words(*start_on_install),
                )
            )
            lines.append("    fi")
        if action_on_upgrade:
            lines.append('    if [ -n "$2" ]; then')
            for action, units in action_on_upgrade.items():
                lines.append(
                    "        deb-systemd-invoke {ACTION} {UNITFILES} >/dev/null || true".format(
                        ACTION=action,
                        UNITFILES=ctrl.maintscript.escape_shell_words(*units),
                    )
                )
            lines.append("    fi")
        lines.append("fi")
        combined = "".join(x if x.endswith("\n") else f"{x}\n" for x in lines)
        ctrl.maintscript.on_configure(combined)

    if stop_then_start_scripts:
        ctrl.maintscript.unconditionally_in_script(
            "preinst",
            textwrap.dedent(
                """\
            if {EMPTY_DPKG_ROOT_CONDITION} && [ "$1" = upgrade ] && {SERVICE_MANAGER_IS_SYSTEMD_CONDITION} ; then
                deb-systemd-invoke stop {UNIT_FILES} >/dev/null || true
            fi
            """.format(
                    EMPTY_DPKG_ROOT_CONDITION=EMPTY_DPKG_ROOT_CONDITION,
                    SERVICE_MANAGER_IS_SYSTEMD_CONDITION=SERVICE_MANAGER_IS_SYSTEMD_CONDITION,
                    UNIT_FILES=ctrl.maintscript.escape_shell_words(
                        *stop_then_start_scripts
                    ),
                )
            ),
        )

    if stop_before_upgrade:
        ctrl.maintscript.on_before_removal(
            """\
        if {EMPTY_DPKG_ROOT_CONDITION} && {SERVICE_MANAGER_IS_SYSTEMD_CONDITION} ; then
            deb-systemd-invoke stop {UNIT_FILES} >/dev/null || true
        fi
        """.format(
                EMPTY_DPKG_ROOT_CONDITION=EMPTY_DPKG_ROOT_CONDITION,
                SERVICE_MANAGER_IS_SYSTEMD_CONDITION=SERVICE_MANAGER_IS_SYSTEMD_CONDITION,
                UNIT_FILES=ctrl.maintscript.escape_shell_words(*stop_before_upgrade),
            )
        )
    if on_purge:
        ctrl.maintscript.on_purge(
            """\
        if [ -x "/usr/bin/deb-systemd-helper" ]; then
            deb-systemd-helper purge {UNITFILES} >/dev/null || true
        fi
        """.format(
                UNITFILES=ctrl.maintscript.escape_shell_words(*stop_before_upgrade),
            )
        )
    ctrl.maintscript.on_removed(
        textwrap.dedent(
            """\
        if {SERVICE_MANAGER_IS_SYSTEMD_CONDITION} ; then
            systemctl --system daemon-reload >/dev/null || true
        fi
        """.format(
                SERVICE_MANAGER_IS_SYSTEMD_CONDITION=SERVICE_MANAGER_IS_SYSTEMD_CONDITION
            )
        )
    )


def _remove_quote(v: str) -> str:
    if v and v[0] == v[-1] and v[0] in ('"', "'"):
        return v[1:-1]
    return v


def _find_and_analyze_systemd_service_files(
    pkg: BinaryPackage,
    fs_root: VirtualPath,
    systemd_service_dir: Literal["system", "user"],
) -> Iterable[SystemdUnit]:
    service_dirs = [
        f"./usr/lib/systemd/{systemd_service_dir}",
        f"./lib/systemd/{systemd_service_dir}",
    ]
    had_install_sections = set()
    aliases: Dict[str, List[str]] = collections.defaultdict(list)
    seen = set()
    all_files = []
    expected_units = set()
    expected_units_required_by = collections.defaultdict(list)

    for d in service_dirs:
        system_dir = fs_root.lookup(d)
        if not system_dir:
            continue
        for child in system_dir.iterdir:
            if child.is_symlink:
                dest = os.path.basename(child.readlink())
                aliases[dest].append(child.name)
            elif child.is_file and child.name not in seen:
                seen.add(child.name)
                all_files.append(child)
                if "@" in child.name:
                    # dh_installsystemd does not check the contents of templated services,
                    # and we match that.
                    continue
                with child.open() as fd:
                    for line in fd:
                        line = line.strip()
                        line_lc = line.lower()
                        if line_lc == "[install]":
                            had_install_sections.add(child.name)
                        elif line_lc.startswith("alias="):
                            # This code assumes service names cannot contain spaces (as in
                            # if you copy-paste it for another field it might not work)
                            aliases[child.name].extend(
                                _remove_quote(x) for x in line[6:].split()
                            )
                        elif line_lc.startswith("also="):
                            # This code assumes service names cannot contain spaces (as in
                            # if you copy-paste it for another field it might not work)
                            for unit in (_remove_quote(x) for x in line[5:].split()):
                                expected_units_required_by[unit].append(child.absolute)
                                expected_units.add(unit)
    for path in all_files:
        if "@" in path.name:
            # Match dh_installsystemd, which skips templated services
            continue
        names = aliases[path.name]
        _, type_of_service = path.name.rsplit(".", 1)
        expected_units.difference_update(names)
        expected_units.discard(path.name)
        names.extend(x[:-8] for x in list(names) if x.endswith(".service"))
        names.insert(0, path.name)
        if path.name.endswith(".service"):
            names.insert(1, path.name[:-8])
        yield SystemdUnit(
            path,
            names,
            type_of_service,
            systemd_service_dir,
            # Bug (?) compat with dh_installsystemd. All units are started, but only
            # enable those with an `[Install]` section.
            # Possibly related bug #1055599
            enable_by_default=path.name in had_install_sections,
            start_by_default=True,
            had_install_section=path.name in had_install_sections,
        )

    if expected_units:
        for unit_name in expected_units:
            required_by = expected_units_required_by[unit_name]
            required_names = ", ".join(required_by)
            _error(
                f"The unit {unit_name} was required by {required_names} (via Also=...)"
                f" but was not present in the package {pkg.name}"
            )


def generate_snippets_for_init_scripts(
    services: Sequence[ServiceDefinition[None]],
    ctrl: BinaryCtrlAccessor,
    _context: PackageProcessingContext,
) -> None:
    for service_def in services:
        script_name = service_def.path.name
        script_installed_path = service_def.path.absolute

        update_rcd_params = (
            "defaults" if service_def.auto_enable_on_install else "defaults-disabled"
        )

        ctrl.maintscript.unconditionally_in_script(
            "preinst",
            textwrap.dedent(
                """\
            if [ "$1" = "install" ] && [ -n "$2" ] && [ -x {DPKG_ROOT}{SCRIPT_PATH} ] ; then
                chmod +x {DPKG_ROOT}{SCRIPT_PATH} >/dev/null || true
            fi
               """.format(
                    DPKG_ROOT=DPKG_ROOT,
                    SCRIPT_PATH=ctrl.maintscript.escape_shell_words(
                        script_installed_path
                    ),
                )
            ),
        )

        lines = [
            "if {EMPTY_DPKG_ROOT_CONDITION} && [ -x {SCRIPT_PATH} ]; then",
            "    update-rc.d {SCRIPT_NAME} {UPDATE_RCD_PARAMS} >/dev/null || exit 1",
        ]

        if (
            service_def.auto_start_on_install
            and service_def.on_upgrade != "stop-then-start"
        ):
            lines.append('    if [ -z "$2" ]; then')
            lines.append(
                "        invoke-rc.d --skip-systemd-native {SCRIPT_NAME} start >/dev/null || exit 1".format(
                    SCRIPT_NAME=ctrl.maintscript.escape_shell_words(script_name),
                )
            )
            lines.append("    fi")

        if service_def.on_upgrade in ("restart", "reload"):
            lines.append('    if [ -n "$2" ]; then')
            lines.append(
                "        invoke-rc.d --skip-systemd-native {SCRIPT_NAME} {ACTION} >/dev/null || exit 1".format(
                    SCRIPT_NAME=ctrl.maintscript.escape_shell_words(script_name),
                    ACTION=service_def.on_upgrade,
                )
            )
            lines.append("    fi")
        elif service_def.on_upgrade == "stop-then-start":
            lines.append(
                "    invoke-rc.d --skip-systemd-native {SCRIPT_NAME} start >/dev/null || exit 1".format(
                    SCRIPT_NAME=ctrl.maintscript.escape_shell_words(script_name),
                )
            )
            ctrl.maintscript.unconditionally_in_script(
                "preinst",
                textwrap.dedent(
                    """\
                if {EMPTY_DPKG_ROOT_CONDITION} && [ "$1" = "upgrade" ] && [ -x {SCRIPT_PATH} ]; then
                    invoke-rc.d --skip-systemd-native {SCRIPT_NAME} stop > /dev/null || true
                fi
                """.format(
                        EMPTY_DPKG_ROOT_CONDITION=EMPTY_DPKG_ROOT_CONDITION,
                        SCRIPT_PATH=ctrl.maintscript.escape_shell_words(
                            script_installed_path
                        ),
                        SCRIPT_NAME=ctrl.maintscript.escape_shell_words(script_name),
                    )
                ),
            )
        elif service_def.on_upgrade != "do-nothing":
            raise AssertionError(
                f"Missing support for on_upgrade rule: {service_def.on_upgrade}"
            )

        lines.append("fi")
        combined = "".join(x if x.endswith("\n") else f"{x}\n" for x in lines)
        ctrl.maintscript.on_configure(
            combined.format(
                EMPTY_DPKG_ROOT_CONDITION=EMPTY_DPKG_ROOT_CONDITION,
                DPKG_ROOT=DPKG_ROOT,
                UPDATE_RCD_PARAMS=update_rcd_params,
                SCRIPT_PATH=ctrl.maintscript.escape_shell_words(script_installed_path),
                SCRIPT_NAME=ctrl.maintscript.escape_shell_words(script_name),
            )
        )

        ctrl.maintscript.on_removed(
            textwrap.dedent(
                """\
            if [ -x {DPKG_ROOT}{SCRIPT_PATH} ]; then
                chmod -x {DPKG_ROOT}{SCRIPT_PATH} > /dev/null || true
            fi
            """.format(
                    DPKG_ROOT=DPKG_ROOT,
                    SCRIPT_PATH=ctrl.maintscript.escape_shell_words(
                        script_installed_path
                    ),
                )
            )
        )
        ctrl.maintscript.on_purge(
            textwrap.dedent(
                """\
            if {EMPTY_DPKG_ROOT_CONDITION} ; then
                update-rc.d {SCRIPT_NAME} remove >/dev/null
            fi
            """.format(
                    SCRIPT_NAME=ctrl.maintscript.escape_shell_words(script_name),
                    EMPTY_DPKG_ROOT_CONDITION=EMPTY_DPKG_ROOT_CONDITION,
                )
            )
        )


def detect_sysv_init_service_files(
    fs_root: VirtualPath,
    service_registry: ServiceRegistry[None],
    _context: PackageProcessingContext,
) -> None:
    etc_init = fs_root.lookup("/etc/init.d")
    if not etc_init:
        return
    for path in etc_init.iterdir:
        if path.is_dir or not path.is_executable:
            continue

        service_registry.register_service(
            path,
            path.name,
        )



import dataclasses
import functools
import os
import textwrap
from contextlib import suppress
from dataclasses import dataclass, field
from typing import (
    List,
    Dict,
    Iterable,
    Mapping,
    Any,
    Union,
    Optional,
    TypeVar,
    Generic,
    cast,
    Set,
    Tuple,
    Sequence,
    FrozenSet,
)

from debian.debian_support import DpkgArchTable

from debputy.dh.debhelper_emulation import (
    dhe_dbgsym_root_dir,
    assert_no_dbgsym_migration,
    read_dbgsym_file,
)
from ._deb_options_profiles import DebBuildOptionsAndProfiles
from ._manifest_constants import *
from .architecture_support import DpkgArchitectureBuildProcessValuesTable
from .builtin_manifest_rules import builtin_mode_normalization_rules
from .exceptions import (
    DebputySubstitutionError,
    DebputyRuntimeError,
    DebputyRuntimeErrorWithPreamble,
)
from .filesystem_scan import FSPath, FSRootDir, FSROOverlay, FSControlRootDir
from .installations import (
    InstallRule,
    SourcePathMatcher,
    PathAlreadyInstalledOrDiscardedError,
    NoMatchForInstallPatternError,
    InstallRuleContext,
    BinaryPackageInstallRuleContext,
    InstallSearchDirContext,
    SearchDir,
)
from .intermediate_manifest import TarMember, PathType, IntermediateManifest
from .maintscript_snippet import (
    DpkgMaintscriptHelperCommand,
    MaintscriptSnippetContainer,
)
from .manifest_conditions import ConditionContext
from .manifest_parser.base_types import (
    FileSystemMatchRule,
    FileSystemExactMatchRule,
    BuildEnvironments,
)
from .manifest_parser.util import AttributePath
from .packager_provided_files import PackagerProvidedFile
from .packages import BinaryPackage, SourcePackage
from .plugin.api.feature_set import PluginProvidedFeatureSet
from .plugin.api.impl import BinaryCtrlAccessorProviderCreator
from .plugin.api.impl_types import (
    PackageProcessingContextProvider,
    PackageDataTable,
)
from .plugin.api.spec import (
    FlushableSubstvars,
    VirtualPath,
    DebputyIntegrationMode,
    INTEGRATION_MODE_DH_DEBPUTY_RRR,
    INTEGRATION_MODE_FULL,
)
from .plugin.debputy.binary_package_rules import ServiceRule
from .plugin.debputy.to_be_api_types import BuildRule
from .plugin.plugin_state import run_in_context_of_plugin
from .substitution import Substitution
from .transformation_rules import (
    TransformationRule,
    ModeNormalizationTransformationRule,
    NormalizeShebangLineTransformation,
)
from .util import (
    _error,
    _warn,
    debian_policy_normalize_symlink_target,
    generated_content_dir,
    _info,
)
from .yaml import MANIFEST_YAML
from .yaml.compat import CommentedMap, CommentedSeq


class PathNotCoveredByInstallRulesError(DebputyRuntimeErrorWithPreamble):

    @property
    def unmatched_paths(self) -> Sequence[VirtualPath]:
        return self.args[1]

    @property
    def search_dir(self) -> VirtualPath:
        return self.args[2]

    def render_preamble(self) -> None:
        _warn(
            f"The following paths were present in {self.search_dir.fs_path}, but not installed (nor explicitly discarded)."
        )
        _warn("")
        for entry in self.unmatched_paths:
            desc = _describe_missing_path(entry)
            _warn(f" * {desc}")
        _warn("")


@dataclass(slots=True)
class DbgsymInfo:
    binary_package: BinaryPackage
    dbgsym_fs_root: FSPath
    _dbgsym_root_fs: Optional[str]
    dbgsym_ids: List[str]

    @property
    def dbgsym_root_dir(self) -> str:
        root_dir = self._dbgsym_root_fs
        if root_dir is None:
            root_dir = generated_content_dir(
                package=self.binary_package,
                subdir_key="dbgsym-fs-root",
            )
            self._dbgsym_root_fs = root_dir
        return root_dir

    @property
    def dbgsym_ctrl_dir(self) -> FSControlRootDir:
        return FSControlRootDir.create_root_dir(
            os.path.join(self.dbgsym_root_dir, "DEBIAN")
        )


@dataclass(slots=True, frozen=True)
class BinaryPackageData:
    source_package: SourcePackage
    binary_package: BinaryPackage
    binary_staging_root_dir: str
    fs_root: FSPath
    substvars: FlushableSubstvars
    package_metadata_context: PackageProcessingContextProvider
    ctrl_creator: BinaryCtrlAccessorProviderCreator
    dbgsym_info: DbgsymInfo

    @property
    def control_output_dir(self) -> FSControlRootDir:
        return FSControlRootDir.create_root_dir(
            generated_content_dir(
                package=self.binary_package,
                subdir_key="DEBIAN",
            )
        )


@dataclass(slots=True)
class PackageTransformationDefinition:
    binary_package: BinaryPackage
    substitution: Substitution
    is_auto_generated_package: bool
    binary_version: Optional[str] = None
    search_dirs: Optional[List[FileSystemExactMatchRule]] = None
    dpkg_maintscript_helper_snippets: List[DpkgMaintscriptHelperCommand] = field(
        default_factory=list
    )
    maintscript_snippets: Dict[str, MaintscriptSnippetContainer] = field(
        default_factory=dict
    )
    transformations: List[TransformationRule] = field(default_factory=list)
    reserved_packager_provided_files: Dict[str, List[PackagerProvidedFile]] = field(
        default_factory=dict
    )
    install_rules: List[InstallRule] = field(default_factory=list)
    requested_service_rules: List[ServiceRule] = field(default_factory=list)


def _path_to_tar_member(
    path: FSPath,
    clamp_mtime_to: int,
) -> TarMember:
    mtime = float(clamp_mtime_to)
    owner, uid, group, gid = path.tar_owner_info
    mode = path.mode

    if path.has_fs_path:
        mtime = min(mtime, path.mtime)

    if path.is_dir:
        path_type = PathType.DIRECTORY
    elif path.is_file:
        # TODO: someday we will need to deal with hardlinks and it might appear here.
        path_type = PathType.FILE
    elif path.is_symlink:
        # Special-case that we resolve immediately (since we need to normalize the target anyway)
        link_target = debian_policy_normalize_symlink_target(
            path.path,
            path.readlink(),
        )
        return TarMember.virtual_path(
            path.tar_path,
            PathType.SYMLINK,
            mtime,
            link_target=link_target,
            # Force mode to be 0777 as that is the mode we see in the data.tar.  In theory, tar lets you set
            # it to whatever. However, for reproducibility, we have to be well-behaved - and that is 0777.
            mode=0o0777,
            owner=owner,
            uid=uid,
            group=group,
            gid=gid,
        )
    else:
        assert not path.is_symlink
        raise AssertionError(
            f"Unsupported file type: {path.path}  - not a file, dir nor a symlink!"
        )

    if not path.has_fs_path:
        assert not path.is_file
        return TarMember.virtual_path(
            path.tar_path,
            path_type,
            mtime,
            mode=mode,
            owner=owner,
            uid=uid,
            group=group,
            gid=gid,
        )
    may_steal_fs_path = path._can_replace_inline
    return TarMember.from_file(
        path.tar_path,
        path.fs_path,
        mode=mode,
        uid=uid,
        owner=owner,
        gid=gid,
        group=group,
        path_type=path_type,
        path_mtime=mtime,
        clamp_mtime_to=clamp_mtime_to,
        may_steal_fs_path=may_steal_fs_path,
    )


def _generate_intermediate_manifest(
    fs_root: FSPath,
    clamp_mtime_to: int,
) -> Iterable[TarMember]:
    symlinks = []
    for path in fs_root.all_paths():
        tar_member = _path_to_tar_member(path, clamp_mtime_to)
        if tar_member.path_type == PathType.SYMLINK:
            symlinks.append(tar_member)
            continue
        yield tar_member
    yield from symlinks


ST = TypeVar("ST")
T = TypeVar("T")


class AbstractYAMLSubStore(Generic[ST]):
    def __init__(
        self,
        parent_store: Any,
        parent_key: Optional[Union[int, str]],
        store: Optional[ST] = None,
    ) -> None:
        if parent_store is not None and parent_key is not None:
            try:
                from_parent_store = parent_store[parent_key]
            except (KeyError, IndexError):
                from_parent_store = None
            if (
                store is not None
                and from_parent_store is not None
                and store is not parent_store
            ):
                raise ValueError(
                    "Store is provided but is not the one already in the parent store"
                )
            if store is None:
                store = from_parent_store
        self._parent_store = parent_store
        self._parent_key = parent_key
        self._is_detached = (
            parent_key is None or parent_store is None or parent_key not in parent_store
        )
        assert self._is_detached or store is not None
        if store is None:
            store = self._create_new_instance()
        self._store: ST = store

    def _create_new_instance(self) -> ST:
        raise NotImplementedError

    def create_definition_if_missing(self) -> None:
        if self._is_detached:
            self.create_definition()

    def create_definition(self) -> None:
        if not self._is_detached:
            raise RuntimeError("Definition is already present")
        parent_store = self._parent_store
        if parent_store is None:
            raise RuntimeError(
                f"Definition is not attached to any parent!? ({self.__class__.__name__})"
            )
        if isinstance(parent_store, list):
            assert self._parent_key is None
            self._parent_key = len(parent_store)
            self._parent_store.append(self._store)
        else:
            parent_store[self._parent_key] = self._store
        self._is_detached = False

    def remove_definition(self) -> None:
        self._ensure_attached()
        del self._parent_store[self._parent_key]
        if isinstance(self._parent_store, list):
            self._parent_key = None
        self._is_detached = True

    def _ensure_attached(self) -> None:
        if self._is_detached:
            raise RuntimeError("The definition has been removed!")


class AbstractYAMLListSubStore(Generic[T], AbstractYAMLSubStore[List[T]]):
    def _create_new_instance(self) -> List[T]:
        return CommentedSeq()


class AbstractYAMLDictSubStore(Generic[T], AbstractYAMLSubStore[Dict[str, T]]):
    def _create_new_instance(self) -> Dict[str, T]:
        return CommentedMap()


class MutableCondition:
    @classmethod
    def arch_matches(cls, arch_filter: str) -> CommentedMap:
        return CommentedMap({MK_CONDITION_ARCH_MATCHES: arch_filter})

    @classmethod
    def build_profiles_matches(cls, build_profiles_matches: str) -> CommentedMap:
        return CommentedMap(
            {MK_CONDITION_BUILD_PROFILES_MATCHES: build_profiles_matches}
        )


class MutableYAMLSymlink(AbstractYAMLDictSubStore[Any]):
    @classmethod
    def new_symlink(
        cls, link_path: str, link_target: str, condition: Optional[Any]
    ) -> "MutableYAMLSymlink":
        inner = {
            MK_TRANSFORMATIONS_CREATE_SYMLINK_LINK_PATH: link_path,
            MK_TRANSFORMATIONS_CREATE_SYMLINK_LINK_TARGET: link_target,
        }
        content = {MK_TRANSFORMATIONS_CREATE_SYMLINK: inner}
        if condition is not None:
            inner["when"] = condition
        return cls(None, None, store=CommentedMap(content))

    @property
    def symlink_path(self) -> str:
        return self._store[MK_TRANSFORMATIONS_CREATE_SYMLINK][
            MK_TRANSFORMATIONS_CREATE_SYMLINK_LINK_PATH
        ]

    @symlink_path.setter
    def symlink_path(self, path: str) -> None:
        self._store[MK_TRANSFORMATIONS_CREATE_SYMLINK][
            MK_TRANSFORMATIONS_CREATE_SYMLINK_LINK_PATH
        ] = path

    @property
    def symlink_target(self) -> Optional[str]:
        return self._store[MK_TRANSFORMATIONS_CREATE_SYMLINK][
            MK_TRANSFORMATIONS_CREATE_SYMLINK_LINK_TARGET
        ]

    @symlink_target.setter
    def symlink_target(self, target: str) -> None:
        self._store[MK_TRANSFORMATIONS_CREATE_SYMLINK][
            MK_TRANSFORMATIONS_CREATE_SYMLINK_LINK_TARGET
        ] = target


class MutableYAMLConffileManagementItem(AbstractYAMLDictSubStore[Any]):
    @classmethod
    def rm_conffile(
        cls,
        conffile: str,
        prior_to_version: Optional[str],
        owning_package: Optional[str],
    ) -> "MutableYAMLConffileManagementItem":
        r = cls(
            None,
            None,
            store=CommentedMap(
                {
                    MK_CONFFILE_MANAGEMENT_REMOVE: CommentedMap(
                        {MK_CONFFILE_MANAGEMENT_REMOVE_PATH: conffile}
                    )
                }
            ),
        )
        r.prior_to_version = prior_to_version
        r.owning_package = owning_package
        return r

    @classmethod
    def mv_conffile(
        cls,
        old_conffile: str,
        new_conffile: str,
        prior_to_version: Optional[str],
        owning_package: Optional[str],
    ) -> "MutableYAMLConffileManagementItem":
        r = cls(
            None,
            None,
            store=CommentedMap(
                {
                    MK_CONFFILE_MANAGEMENT_RENAME: CommentedMap(
                        {
                            MK_CONFFILE_MANAGEMENT_RENAME_SOURCE: old_conffile,
                            MK_CONFFILE_MANAGEMENT_RENAME_TARGET: new_conffile,
                        }
                    )
                }
            ),
        )
        r.prior_to_version = prior_to_version
        r.owning_package = owning_package
        return r

    @property
    def _container(self) -> Dict[str, Any]:
        assert len(self._store) == 1
        return next(iter(self._store.values()))

    @property
    def command(self) -> str:
        assert len(self._store) == 1
        return next(iter(self._store))

    @property
    def obsolete_conffile(self) -> str:
        if self.command == MK_CONFFILE_MANAGEMENT_REMOVE:
            return self._container[MK_CONFFILE_MANAGEMENT_REMOVE_PATH]
        assert self.command == MK_CONFFILE_MANAGEMENT_RENAME
        return self._container[MK_CONFFILE_MANAGEMENT_RENAME_SOURCE]

    @obsolete_conffile.setter
    def obsolete_conffile(self, value: str) -> None:
        if self.command == MK_CONFFILE_MANAGEMENT_REMOVE:
            self._container[MK_CONFFILE_MANAGEMENT_REMOVE_PATH] = value
        else:
            assert self.command == MK_CONFFILE_MANAGEMENT_RENAME
            self._container[MK_CONFFILE_MANAGEMENT_RENAME_SOURCE] = value

    @property
    def new_conffile(self) -> str:
        if self.command != MK_CONFFILE_MANAGEMENT_RENAME:
            raise TypeError(
                f"The new_conffile attribute is only applicable to command {MK_CONFFILE_MANAGEMENT_RENAME}."
                f" This is a {self.command}"
            )
        return self._container[MK_CONFFILE_MANAGEMENT_RENAME_TARGET]

    @new_conffile.setter
    def new_conffile(self, value: str) -> None:
        if self.command != MK_CONFFILE_MANAGEMENT_RENAME:
            raise TypeError(
                f"The new_conffile attribute is only applicable to command {MK_CONFFILE_MANAGEMENT_RENAME}."
                f" This is a {self.command}"
            )
        self._container[MK_CONFFILE_MANAGEMENT_RENAME_TARGET] = value

    @property
    def prior_to_version(self) -> Optional[str]:
        return self._container.get(MK_CONFFILE_MANAGEMENT_X_PRIOR_TO_VERSION)

    @prior_to_version.setter
    def prior_to_version(self, value: Optional[str]) -> None:
        if value is None:
            try:
                del self._container[MK_CONFFILE_MANAGEMENT_X_PRIOR_TO_VERSION]
            except KeyError:
                pass
        else:
            self._container[MK_CONFFILE_MANAGEMENT_X_PRIOR_TO_VERSION] = value

    @property
    def owning_package(self) -> Optional[str]:
        return self._container[MK_CONFFILE_MANAGEMENT_X_PRIOR_TO_VERSION]

    @owning_package.setter
    def owning_package(self, value: Optional[str]) -> None:
        if value is None:
            try:
                del self._container[MK_CONFFILE_MANAGEMENT_X_OWNING_PACKAGE]
            except KeyError:
                pass
        else:
            self._container[MK_CONFFILE_MANAGEMENT_X_OWNING_PACKAGE] = value


class MutableYAMLPackageDefinition(AbstractYAMLDictSubStore):
    def _list_store(
        self, key, *, create_if_absent: bool = False
    ) -> Optional[List[Dict[str, Any]]]:
        if self._is_detached or key not in self._store:
            if create_if_absent:
                return None
            self.create_definition_if_missing()
            self._store[key] = []
        return self._store[key]

    def _insert_item(self, key: str, item: AbstractYAMLDictSubStore) -> None:
        parent_store = self._list_store(key, create_if_absent=True)
        assert parent_store is not None
        if not item._is_detached or (
            item._parent_store is not None and item._parent_store is not parent_store
        ):
            raise RuntimeError(
                "Item is already attached or associated with a different container"
            )
        item._parent_store = parent_store
        item.create_definition()

    def add_symlink(self, symlink: MutableYAMLSymlink) -> None:
        self._insert_item(MK_TRANSFORMATIONS, symlink)

    def symlinks(self) -> Iterable[MutableYAMLSymlink]:
        store = self._list_store(MK_TRANSFORMATIONS)
        if store is None:
            return
        for i in range(len(store)):
            d = store[i]
            if d and isinstance(d, dict) and len(d) == 1 and "symlink" in d:
                yield MutableYAMLSymlink(store, i)

    def conffile_management_items(self) -> Iterable[MutableYAMLConffileManagementItem]:
        store = self._list_store(MK_CONFFILE_MANAGEMENT)
        if store is None:
            return
        yield from (
            MutableYAMLConffileManagementItem(store, i) for i in range(len(store))
        )

    def add_conffile_management(
        self, conffile_management_item: MutableYAMLConffileManagementItem
    ) -> None:
        self._insert_item(MK_CONFFILE_MANAGEMENT, conffile_management_item)


class AbstractMutableYAMLInstallRule(AbstractYAMLDictSubStore):
    @property
    def _container(self) -> Dict[str, Any]:
        assert len(self._store) == 1
        return next(iter(self._store.values()))

    @property
    def into(self) -> Optional[List[str]]:
        v = self._container[MK_INSTALLATIONS_INSTALL_INTO]
        if v is None:
            return None
        if isinstance(v, str):
            return [v]
        return v

    @into.setter
    def into(self, new_value: Optional[Union[str, List[str]]]) -> None:
        if new_value is None:
            with suppress(KeyError):
                del self._container[MK_INSTALLATIONS_INSTALL_INTO]
            return
        if isinstance(new_value, str):
            self._container[MK_INSTALLATIONS_INSTALL_INTO] = new_value
            return
        new_list = CommentedSeq(new_value)
        self._container[MK_INSTALLATIONS_INSTALL_INTO] = new_list

    @property
    def when(self) -> Optional[Union[str, Mapping[str, Any]]]:
        return self._container[MK_CONDITION_WHEN]

    @when.setter
    def when(self, new_value: Optional[Union[str, Mapping[str, Any]]]) -> None:
        if new_value is None:
            with suppress(KeyError):
                del self._container[MK_CONDITION_WHEN]
            return
        if isinstance(new_value, str):
            self._container[MK_CONDITION_WHEN] = new_value
            return
        new_map = CommentedMap(new_value)
        self._container[MK_CONDITION_WHEN] = new_map

    @classmethod
    def install_dest(
        cls,
        sources: Union[str, List[str]],
        into: Optional[Union[str, List[str]]],
        *,
        dest_dir: Optional[str] = None,
        when: Optional[Union[str, Mapping[str, Any]]] = None,
    ) -> "MutableYAMLInstallRuleInstall":
        k = MK_INSTALLATIONS_INSTALL_SOURCES
        if isinstance(sources, str):
            k = MK_INSTALLATIONS_INSTALL_SOURCE
        r = MutableYAMLInstallRuleInstall(
            None,
            None,
            store=CommentedMap(
                {
                    MK_INSTALLATIONS_INSTALL: CommentedMap(
                        {
                            k: sources,
                        }
                    )
                }
            ),
        )
        r.dest_dir = dest_dir
        r.into = into
        if when is not None:
            r.when = when
        return r

    @classmethod
    def multi_dest_install(
        cls,
        sources: Union[str, List[str]],
        dest_dirs: Sequence[str],
        into: Optional[Union[str, List[str]]],
        *,
        when: Optional[Union[str, Mapping[str, Any]]] = None,
    ) -> "MutableYAMLInstallRuleInstall":
        k = MK_INSTALLATIONS_INSTALL_SOURCES
        if isinstance(sources, str):
            k = MK_INSTALLATIONS_INSTALL_SOURCE
        r = MutableYAMLInstallRuleInstall(
            None,
            None,
            store=CommentedMap(
                {
                    MK_INSTALLATIONS_MULTI_DEST_INSTALL: CommentedMap(
                        {
                            k: sources,
                            "dest-dirs": dest_dirs,
                        }
                    )
                }
            ),
        )
        r.into = into
        if when is not None:
            r.when = when
        return r

    @classmethod
    def install_as(
        cls,
        source: str,
        install_as: str,
        into: Optional[Union[str, List[str]]],
        when: Optional[Union[str, Mapping[str, Any]]] = None,
    ) -> "MutableYAMLInstallRuleInstall":
        r = MutableYAMLInstallRuleInstall(
            None,
            None,
            store=CommentedMap(
                {
                    MK_INSTALLATIONS_INSTALL: CommentedMap(
                        {
                            MK_INSTALLATIONS_INSTALL_SOURCE: source,
                            MK_INSTALLATIONS_INSTALL_AS: install_as,
                        }
                    )
                }
            ),
        )
        r.into = into
        if when is not None:
            r.when = when
        return r

    @classmethod
    def install_doc_as(
        cls,
        source: str,
        install_as: str,
        into: Optional[Union[str, List[str]]],
        when: Optional[Union[str, Mapping[str, Any]]] = None,
    ) -> "MutableYAMLInstallRuleInstall":
        r = MutableYAMLInstallRuleInstall(
            None,
            None,
            store=CommentedMap(
                {
                    MK_INSTALLATIONS_INSTALL_DOCS: CommentedMap(
                        {
                            MK_INSTALLATIONS_INSTALL_SOURCE: source,
                            MK_INSTALLATIONS_INSTALL_AS: install_as,
                        }
                    )
                }
            ),
        )
        r.into = into
        if when is not None:
            r.when = when
        return r

    @classmethod
    def install_docs(
        cls,
        sources: Union[str, List[str]],
        into: Optional[Union[str, List[str]]],
        *,
        dest_dir: Optional[str] = None,
        when: Optional[Union[str, Mapping[str, Any]]] = None,
    ) -> "MutableYAMLInstallRuleInstall":
        k = MK_INSTALLATIONS_INSTALL_SOURCES
        if isinstance(sources, str):
            k = MK_INSTALLATIONS_INSTALL_SOURCE
        r = MutableYAMLInstallRuleInstall(
            None,
            None,
            store=CommentedMap(
                {
                    MK_INSTALLATIONS_INSTALL_DOCS: CommentedMap(
                        {
                            k: sources,
                        }
                    )
                }
            ),
        )
        r.into = into
        r.dest_dir = dest_dir
        if when is not None:
            r.when = when
        return r

    @classmethod
    def install_examples(
        cls,
        sources: Union[str, List[str]],
        into: Optional[Union[str, List[str]]],
        when: Optional[Union[str, Mapping[str, Any]]] = None,
    ) -> "MutableYAMLInstallRuleInstallExamples":
        k = MK_INSTALLATIONS_INSTALL_SOURCES
        if isinstance(sources, str):
            k = MK_INSTALLATIONS_INSTALL_SOURCE
        r = MutableYAMLInstallRuleInstallExamples(
            None,
            None,
            store=CommentedMap(
                {
                    MK_INSTALLATIONS_INSTALL_EXAMPLES: CommentedMap(
                        {
                            k: sources,
                        }
                    )
                }
            ),
        )
        r.into = into
        if when is not None:
            r.when = when
        return r

    @classmethod
    def install_man(
        cls,
        sources: Union[str, List[str]],
        into: Optional[Union[str, List[str]]],
        language: Optional[str],
        when: Optional[Union[str, Mapping[str, Any]]] = None,
    ) -> "MutableYAMLInstallRuleMan":
        k = MK_INSTALLATIONS_INSTALL_SOURCES
        if isinstance(sources, str):
            k = MK_INSTALLATIONS_INSTALL_SOURCE
        r = MutableYAMLInstallRuleMan(
            None,
            None,
            store=CommentedMap(
                {
                    MK_INSTALLATIONS_INSTALL_MAN: CommentedMap(
                        {
                            k: sources,
                        }
                    )
                }
            ),
        )
        r.language = language
        r.into = into
        if when is not None:
            r.when = when
        return r

    @classmethod
    def discard(
        cls,
        sources: Union[str, List[str]],
    ) -> "MutableYAMLInstallRuleDiscard":
        return MutableYAMLInstallRuleDiscard(
            None,
            None,
            store=CommentedMap({MK_INSTALLATIONS_DISCARD: sources}),
        )


class MutableYAMLInstallRuleInstallExamples(AbstractMutableYAMLInstallRule):
    pass


class MutableYAMLInstallRuleMan(AbstractMutableYAMLInstallRule):
    @property
    def language(self) -> Optional[str]:
        return self._container[MK_INSTALLATIONS_INSTALL_MAN_LANGUAGE]

    @language.setter
    def language(self, new_value: Optional[str]) -> None:
        if new_value is not None:
            self._container[MK_INSTALLATIONS_INSTALL_MAN_LANGUAGE] = new_value
            return
        with suppress(KeyError):
            del self._container[MK_INSTALLATIONS_INSTALL_MAN_LANGUAGE]


class MutableYAMLInstallRuleDiscard(AbstractMutableYAMLInstallRule):
    pass


class MutableYAMLInstallRuleInstall(AbstractMutableYAMLInstallRule):
    @property
    def sources(self) -> List[str]:
        v = self._container[MK_INSTALLATIONS_INSTALL_SOURCES]
        if isinstance(v, str):
            return [v]
        return v

    @sources.setter
    def sources(self, new_value: Union[str, List[str]]) -> None:
        if isinstance(new_value, str):
            self._container[MK_INSTALLATIONS_INSTALL_SOURCES] = new_value
            return
        new_list = CommentedSeq(new_value)
        self._container[MK_INSTALLATIONS_INSTALL_SOURCES] = new_list

    @property
    def dest_dir(self) -> Optional[str]:
        return self._container.get(MK_INSTALLATIONS_INSTALL_DEST_DIR)

    @dest_dir.setter
    def dest_dir(self, new_value: Optional[str]) -> None:
        if new_value is not None and self.dest_as is not None:
            raise ValueError(
                f'Cannot both have a "{MK_INSTALLATIONS_INSTALL_DEST_DIR}" and'
                f' "{MK_INSTALLATIONS_INSTALL_AS}"'
            )
        if new_value is not None:
            self._container[MK_INSTALLATIONS_INSTALL_DEST_DIR] = new_value
        else:
            with suppress(KeyError):
                del self._container[MK_INSTALLATIONS_INSTALL_DEST_DIR]

    @property
    def dest_as(self) -> Optional[str]:
        return self._container.get(MK_INSTALLATIONS_INSTALL_AS)

    @dest_as.setter
    def dest_as(self, new_value: Optional[str]) -> None:
        if new_value is not None:
            if self.dest_dir is not None:
                raise ValueError(
                    f'Cannot both have a "{MK_INSTALLATIONS_INSTALL_DEST_DIR}" and'
                    f' "{MK_INSTALLATIONS_INSTALL_AS}"'
                )

            sources = self._container[MK_INSTALLATIONS_INSTALL_SOURCES]
            if isinstance(sources, list):
                if len(sources) != 1:
                    raise ValueError(
                        f'Cannot have "{MK_INSTALLATIONS_INSTALL_AS}" when'
                        f' "{MK_INSTALLATIONS_INSTALL_SOURCES}" is not exactly one item'
                    )
                self.sources = sources[0]
            self._container[MK_INSTALLATIONS_INSTALL_AS] = new_value
        else:
            with suppress(KeyError):
                del self._container[MK_INSTALLATIONS_INSTALL_AS]


class MutableYAMLInstallationsDefinition(AbstractYAMLListSubStore[Any]):
    def append(self, install_rule: AbstractMutableYAMLInstallRule) -> None:
        parent_store = self._store
        if not install_rule._is_detached or (
            install_rule._parent_store is not None
            and install_rule._parent_store is not parent_store
        ):
            raise RuntimeError(
                "Item is already attached or associated with a different container"
            )
        self.create_definition_if_missing()
        install_rule._parent_store = parent_store
        install_rule.create_definition()

    def extend(self, install_rules: Iterable[AbstractMutableYAMLInstallRule]) -> None:
        parent_store = self._store
        for install_rule in install_rules:
            if not install_rule._is_detached or (
                install_rule._parent_store is not None
                and install_rule._parent_store is not parent_store
            ):
                raise RuntimeError(
                    "Item is already attached or associated with a different container"
                )
            self.create_definition_if_missing()
            install_rule._parent_store = parent_store
            install_rule.create_definition()


class MutableYAMLManifestVariables(AbstractYAMLDictSubStore):
    @property
    def variables(self) -> Dict[str, Any]:
        return self._store

    def __setitem__(self, key: str, value: Any) -> None:
        self._store[key] = value
        self.create_definition_if_missing()


class MutableYAMLManifestDefinitions(AbstractYAMLDictSubStore):
    def manifest_variables(
        self, *, create_if_absent: bool = True
    ) -> MutableYAMLManifestVariables:
        d = MutableYAMLManifestVariables(self._store, MK_MANIFEST_VARIABLES)
        if create_if_absent:
            d.create_definition_if_missing()
        return d


class MutableYAMLManifest:
    def __init__(self, store: Any) -> None:
        self._store = store

    @classmethod
    def empty_manifest(cls) -> "MutableYAMLManifest":
        return cls(CommentedMap({MK_MANIFEST_VERSION: DEFAULT_MANIFEST_VERSION}))

    @property
    def manifest_version(self) -> str:
        return self._store[MK_MANIFEST_VERSION]

    @manifest_version.setter
    def manifest_version(self, version: str) -> None:
        if version not in SUPPORTED_MANIFEST_VERSIONS:
            raise ValueError("Unsupported version")
        self._store[MK_MANIFEST_VERSION] = version

    def installations(
        self,
        *,
        create_if_absent: bool = True,
    ) -> MutableYAMLInstallationsDefinition:
        d = MutableYAMLInstallationsDefinition(self._store, MK_INSTALLATIONS)
        if create_if_absent:
            d.create_definition_if_missing()
        return d

    def manifest_definitions(
        self,
        *,
        create_if_absent: bool = True,
    ) -> MutableYAMLManifestDefinitions:
        d = MutableYAMLManifestDefinitions(self._store, MK_MANIFEST_DEFINITIONS)
        if create_if_absent:
            d.create_definition_if_missing()
        return d

    def package(
        self, name: str, *, create_if_absent: bool = True
    ) -> MutableYAMLPackageDefinition:
        if MK_PACKAGES not in self._store:
            self._store[MK_PACKAGES] = CommentedMap()
        packages_store = self._store[MK_PACKAGES]
        package = packages_store.get(name)
        if package is None:
            if not create_if_absent:
                raise KeyError(name)
            assert packages_store is not None
            d = MutableYAMLPackageDefinition(packages_store, name)
            d.create_definition()
        else:
            d = MutableYAMLPackageDefinition(packages_store, name)
        return d

    def write_to(self, fd) -> None:
        MANIFEST_YAML.dump(self._store, fd)


def _describe_missing_path(entry: VirtualPath) -> str:
    if entry.is_dir:
        return f"{entry.fs_path}/ (empty directory; possible integration point)"
    if entry.is_symlink:
        target = os.readlink(entry.fs_path)
        return f"{entry.fs_path} (symlink; links to {target})"
    if entry.is_file:
        return f"{entry.fs_path} (file)"
    return f"{entry.fs_path} (other!? Probably not supported by debputy and may need a `remove`)"


def _detect_missing_installations(
    path_matcher: SourcePathMatcher,
    search_dir: VirtualPath,
) -> None:
    if not search_dir.is_dir:
        return
    missing = list(path_matcher.detect_missing(search_dir))
    if not missing:
        return

    excl = textwrap.dedent(
        """\
        - discard: "*"
    """
    )

    raise PathNotCoveredByInstallRulesError(
        "Please review the list and add either install rules or exclusions to `installations` in"
        " debian/debputy.manifest.  If you do not need any of these paths, add the following to the"
        f" end of your 'installations`:\n\n{excl}\n",
        missing,
        search_dir,
    )


def _list_automatic_discard_rules(path_matcher: SourcePathMatcher) -> None:
    used_discard_rules = path_matcher.used_auto_discard_rules
    # Discard rules can match and then be overridden.  In that case, they appear
    # but have 0 matches.
    if not sum((len(v) for v in used_discard_rules.values()), 0):
        return
    _info("The following automatic discard rules were triggered:")
    example_path: Optional[str] = None
    for rule in sorted(used_discard_rules):
        for fs_path in sorted(used_discard_rules[rule]):
            if example_path is None:
                example_path = fs_path
            _info(f" * {rule} -> {fs_path}")
    assert example_path is not None
    _info("")
    _info(
        "Note that some of these may have been overruled. The overrule detection logic is not"
    )
    _info("100% reliable.")
    _info("")
    _info(
        "You can overrule an automatic discard rule by explicitly listing the path. As an example:"
    )
    _info("    installations:")
    _info("    - install:")
    _info(f"        source: {example_path}")


def _install_everything_from_source_dir_if_present(
    dctrl_bin: BinaryPackage,
    substitution: Substitution,
    path_matcher: SourcePathMatcher,
    install_rule_context: InstallRuleContext,
    source_condition_context: ConditionContext,
    source_dir: VirtualPath,
    *,
    into_dir: Optional[VirtualPath] = None,
) -> None:
    attribute_path = AttributePath.builtin_path()[f"installing {source_dir.fs_path}"]
    pkg_set = frozenset([dctrl_bin])
    install_rule = run_in_context_of_plugin(
        "debputy",
        InstallRule.install_dest,
        [FileSystemMatchRule.from_path_match("*", attribute_path, substitution)],
        None,
        pkg_set,
        f"Built-in; install everything from {source_dir.fs_path} into {dctrl_bin.name}",
        None,
    )
    pkg_search_dir: Tuple[SearchDir] = (
        SearchDir(
            source_dir,
            pkg_set,
        ),
    )
    replacements = {
        "search_dirs": pkg_search_dir,
    }
    if into_dir is not None:
        binary_package_contexts = dict(install_rule_context.binary_package_contexts)
        updated = binary_package_contexts[dctrl_bin.name].replace(fs_root=into_dir)
        binary_package_contexts[dctrl_bin.name] = updated
        replacements["binary_package_contexts"] = binary_package_contexts

    fake_install_rule_context = install_rule_context.replace(**replacements)
    try:
        install_rule.perform_install(
            path_matcher,
            fake_install_rule_context,
            source_condition_context,
        )
    except (
        NoMatchForInstallPatternError,
        PathAlreadyInstalledOrDiscardedError,
    ):
        # Empty directory or everything excluded by default; ignore the error
        pass


class HighLevelManifest:
    def __init__(
        self,
        manifest_path: str,
        mutable_manifest: Optional[MutableYAMLManifest],
        install_rules: Optional[List[InstallRule]],
        source_package: SourcePackage,
        binary_packages: Mapping[str, BinaryPackage],
        substitution: Substitution,
        package_transformations: Mapping[str, PackageTransformationDefinition],
        dpkg_architecture_variables: DpkgArchitectureBuildProcessValuesTable,
        dpkg_arch_query_table: DpkgArchTable,
        build_env: DebBuildOptionsAndProfiles,
        build_environments: BuildEnvironments,
        build_rules: Optional[List[BuildRule]],
        plugin_provided_feature_set: PluginProvidedFeatureSet,
        debian_dir: VirtualPath,
    ) -> None:
        self.manifest_path = manifest_path
        self.mutable_manifest = mutable_manifest
        self._install_rules = install_rules
        self._source_package = source_package
        self._binary_packages = binary_packages
        self.substitution = substitution
        self.package_transformations = package_transformations
        self._dpkg_architecture_variables = dpkg_architecture_variables
        self._dpkg_arch_query_table = dpkg_arch_query_table
        self._build_env = build_env
        self._used_for: Set[str] = set()
        self.build_environments = build_environments
        self.build_rules = build_rules
        self._plugin_provided_feature_set = plugin_provided_feature_set
        self._debian_dir = debian_dir
        self._source_condition_context = ConditionContext(
            binary_package=None,
            substitution=self.substitution,
            deb_options_and_profiles=self._build_env,
            dpkg_architecture_variables=self._dpkg_architecture_variables,
            dpkg_arch_query_table=self._dpkg_arch_query_table,
        )

    def source_version(self, include_binnmu_version: bool = True) -> str:
        # TODO: There should an easier way to determine the source version; really.
        version_var = "{{DEB_VERSION}}"
        if not include_binnmu_version:
            version_var = "{{_DEBPUTY_INTERNAL_NON_BINNMU_SOURCE}}"
        try:
            return self.substitution.substitute(
                version_var, "internal (resolve version)"
            )
        except DebputySubstitutionError as e:
            raise AssertionError(f"Could not resolve {version_var}") from e

    @property
    def source_condition_context(self) -> ConditionContext:
        return self._source_condition_context

    @property
    def debian_dir(self) -> VirtualPath:
        return self._debian_dir

    @property
    def dpkg_architecture_variables(self) -> DpkgArchitectureBuildProcessValuesTable:
        return self._dpkg_architecture_variables

    @property
    def deb_options_and_profiles(self) -> DebBuildOptionsAndProfiles:
        return self._build_env

    @property
    def plugin_provided_feature_set(self) -> PluginProvidedFeatureSet:
        return self._plugin_provided_feature_set

    @property
    def active_packages(self) -> Iterable[BinaryPackage]:
        yield from (p for p in self._binary_packages.values() if p.should_be_acted_on)

    @property
    def all_packages(self) -> Iterable[BinaryPackage]:
        yield from self._binary_packages.values()

    def package_state_for(self, package: str) -> PackageTransformationDefinition:
        return self.package_transformations[package]

    def _detect_doc_main_package_for(self, package: BinaryPackage) -> BinaryPackage:
        name = package.name
        # If it is not a -doc package, then docs should be installed
        # under its own package name.
        if not name.endswith("-doc"):
            return package
        name = name[:-4]
        main_package = self._binary_packages.get(name)
        if main_package:
            return main_package
        if name.startswith("lib"):
            dev_pkg = self._binary_packages.get(f"{name}-dev")
            if dev_pkg:
                return dev_pkg

        # If we found no better match; default to the doc package itself.
        return package

    def perform_installations(
        self,
        integration_mode: DebputyIntegrationMode,
        *,
        install_request_context: Optional[InstallSearchDirContext] = None,
    ) -> PackageDataTable:
        package_data_dict = {}
        package_data_table = PackageDataTable(package_data_dict)
        enable_manifest_installation_feature = (
            integration_mode != INTEGRATION_MODE_DH_DEBPUTY_RRR
        )
        if install_request_context is None:

            @functools.lru_cache(None)
            def _as_path(fs_path: str) -> VirtualPath:
                return FSROOverlay.create_root_dir(".", fs_path)

            dtmp_dir = _as_path("debian/tmp")
            source_root_dir = _as_path(".")
            into = frozenset(self._binary_packages.values())
            per_package_search_dirs = {
                t.binary_package: [_as_path(f.match_rule.path) for f in t.search_dirs]
                for t in self.package_transformations.values()
                if t.search_dirs is not None
            }

            if integration_mode == INTEGRATION_MODE_FULL:
                # In this mode, we have no default search dir. Everything ends up being
                # per-package instead (since that is easier logic-wise).
                #
                # Even dtmp_dir is omitted here (since it is not universally applicable).
                # Note we still initialize dtmp_dir, since it affects a later guard.
                default_search_dirs = []
                seen_pp_search_dirs: Set[Tuple[BinaryPackage, str]] = set()
                # TODO: We should **not** re-detect the default build system here. Instead,
                #  we should have accumulated these per-package install dirs when we
                #  invoked the build systems (notably, we are also not using the right
                #  environment, which happens to work for now but that might break in
                #  the future).
                build_rules = self.build_rules
                if build_rules is None:
                    from .build_support.buildsystem_detection import (
                        auto_detect_buildsystem,
                    )

                    bs = auto_detect_buildsystem(self)
                    build_rules = [bs] if bs is not None else []

                for build_rule in build_rules:
                    dest_dir = build_rule.install_dest_dir()
                    if dest_dir is None:
                        continue
                    dest_path = _as_path(dest_dir)
                    for pkg in build_rule.for_packages:
                        seen_key = (pkg, dest_dir)
                        if seen_key in seen_pp_search_dirs:
                            continue
                        seen_pp_search_dirs.add(seen_key)
                        if pkg not in per_package_search_dirs:
                            per_package_search_dirs[pkg] = [dest_path]
                        else:
                            per_package_search_dirs[pkg].append(dest_path)

                # We can end here with per_package_search_dirs having no search dirs for any package
                # (single binary, where everything is installed into d/<pkg> is the most common case).
                #
                # This is not a problem in itself as the installation rules can still apply to the
                # source root and there should be no reason to install something from d/<pkg> into
                # d/<another-pkg>
            else:
                default_search_dirs = [dtmp_dir]

            search_dirs = _determine_search_dir_order(
                per_package_search_dirs,
                into,
                default_search_dirs,
                source_root_dir,
            )
            check_for_uninstalled_dirs = tuple(
                s.search_dir
                for s in search_dirs
                if s.search_dir.fs_path != source_root_dir.fs_path
            )
            if enable_manifest_installation_feature:
                _present_installation_dirs(
                    search_dirs, check_for_uninstalled_dirs, into
                )
        else:
            dtmp_dir = None
            search_dirs = install_request_context.search_dirs
            into = frozenset(self._binary_packages.values())
            seen: Set[BinaryPackage] = set()
            for search_dir in search_dirs:
                seen.update(search_dir.applies_to)

            missing = into - seen
            if missing:
                names = ", ".join(p.name for p in missing)
                raise ValueError(
                    f"The following package(s) had no search dirs: {names}."
                    " (Generally, the source root would be applicable to all packages)"
                )
            extra_names = seen - into
            if extra_names:
                names = ", ".join(p.name for p in extra_names)
                raise ValueError(
                    f"The install_request_context referenced the following unknown package(s): {names}"
                )

            check_for_uninstalled_dirs = (
                install_request_context.check_for_uninstalled_dirs
            )

        install_rule_context = InstallRuleContext(search_dirs)

        if (
            enable_manifest_installation_feature
            and self._install_rules is None
            # TODO: Should we also do this for full mode when build systems provided search dirs?
            and dtmp_dir is not None
            and os.path.isdir(dtmp_dir.fs_path)
        ):
            msg = (
                "The build system appears to have provided the output of upstream build system's"
                " install in debian/tmp.  However, these are no provisions for debputy to install"
                " any of that into any of the debian packages listed in debian/control."
                " To avoid accidentally creating empty packages, debputy will insist that you "
                " explicitly define an empty installation definition if you did not want to "
                " install any of those files even though they have been provided."
                ' Example: "installations: []"'
            )
            _error(msg)
        elif (
            not enable_manifest_installation_feature and self._install_rules is not None
        ):
            _error(
                f"The `installations` feature cannot be used in {self.manifest_path} with this integration mode."
                f" Please remove or comment out the `installations` keyword."
            )

        for dctrl_bin in self.all_packages:
            package = dctrl_bin.name
            doc_main_package = self._detect_doc_main_package_for(dctrl_bin)

            install_rule_context[package] = BinaryPackageInstallRuleContext(
                dctrl_bin,
                FSRootDir(),
                doc_main_package,
            )

        if enable_manifest_installation_feature:
            discard_rules = list(
                self.plugin_provided_feature_set.auto_discard_rules.values()
            )
        else:
            discard_rules = [
                self.plugin_provided_feature_set.auto_discard_rules["debian-dir"]
            ]
        path_matcher = SourcePathMatcher(discard_rules)

        source_condition_context = self._source_condition_context

        for dctrl_bin in self.active_packages:
            package = dctrl_bin.name
            if install_request_context:
                build_system_staging_dir = install_request_context.debian_pkg_dirs.get(
                    package
                )
            else:
                build_system_staging_dir_fs_path = os.path.join("debian", package)
                if os.path.isdir(build_system_staging_dir_fs_path):
                    build_system_staging_dir = FSROOverlay.create_root_dir(
                        ".",
                        build_system_staging_dir_fs_path,
                    )
                else:
                    build_system_staging_dir = None

            if build_system_staging_dir is not None:
                _install_everything_from_source_dir_if_present(
                    dctrl_bin,
                    self.substitution,
                    path_matcher,
                    install_rule_context,
                    source_condition_context,
                    build_system_staging_dir,
                )

        if self._install_rules:
            # FIXME: Check that every install rule remains used after transformations have run.
            # What we want to check is transformations do not exclude everything from an install
            # rule. The hard part here is that renaming (etc.) is fine, so we cannot 1:1 string
            # match.
            for install_rule in self._install_rules:
                install_rule.perform_install(
                    path_matcher,
                    install_rule_context,
                    source_condition_context,
                )

        if enable_manifest_installation_feature:
            for search_dir in check_for_uninstalled_dirs:
                _detect_missing_installations(path_matcher, search_dir)

        for dctrl_bin in self.all_packages:
            package = dctrl_bin.name
            binary_install_rule_context = install_rule_context[package]
            build_system_pkg_staging_dir = os.path.join("debian", package)
            fs_root = binary_install_rule_context.fs_root

            context = self.package_transformations[package]
            if dctrl_bin.should_be_acted_on and enable_manifest_installation_feature:
                for special_install_rule in context.install_rules:
                    special_install_rule.perform_install(
                        path_matcher,
                        install_rule_context,
                        source_condition_context,
                    )

            if dctrl_bin.should_be_acted_on:
                self.apply_fs_transformations(package, fs_root)
                substvars_file = f"debian/{package}.substvars"
                substvars = FlushableSubstvars.load_from_path(
                    substvars_file, missing_ok=True
                )
                # We do not want to touch the substvars file (non-clean rebuild contamination)
                substvars.substvars_path = None
            else:
                substvars = FlushableSubstvars()

            udeb_package = self._binary_packages.get(f"{package}-udeb")
            if udeb_package and not udeb_package.is_udeb:
                udeb_package = None

            package_metadata_context = PackageProcessingContextProvider(
                self,
                dctrl_bin,
                udeb_package,
                package_data_table,
                # FIXME: source_package
            )

            ctrl_creator = BinaryCtrlAccessorProviderCreator(
                package_metadata_context,
                substvars,
                context.maintscript_snippets,
                context.substitution,
            )

            if not enable_manifest_installation_feature:
                assert_no_dbgsym_migration(dctrl_bin)
                dh_dbgsym_root_fs = dhe_dbgsym_root_dir(dctrl_bin)
                dh_dbgsym_root_path = FSROOverlay.create_root_dir(
                    "",
                    dh_dbgsym_root_fs,
                )
                dbgsym_root_fs = FSRootDir()
                _install_everything_from_source_dir_if_present(
                    dctrl_bin,
                    self.substitution,
                    path_matcher,
                    install_rule_context,
                    source_condition_context,
                    dh_dbgsym_root_path,
                    into_dir=dbgsym_root_fs,
                )
                dbgsym_build_ids = read_dbgsym_file(dctrl_bin)
                dbgsym_info = DbgsymInfo(
                    dctrl_bin,
                    dbgsym_root_fs,
                    os.path.join(dh_dbgsym_root_fs, "DEBIAN"),
                    dbgsym_build_ids,
                )
            else:
                dbgsym_info = DbgsymInfo(
                    dctrl_bin,
                    FSRootDir(),
                    None,
                    [],
                )

            package_data_dict[package] = BinaryPackageData(
                self._source_package,
                dctrl_bin,
                build_system_pkg_staging_dir,
                fs_root,
                substvars,
                package_metadata_context,
                ctrl_creator,
                dbgsym_info,
            )

        if enable_manifest_installation_feature:
            _list_automatic_discard_rules(path_matcher)

        return package_data_table

    def condition_context(
        self, binary_package: Optional[Union[BinaryPackage, str]]
    ) -> ConditionContext:
        if binary_package is None:
            return self._source_condition_context
        if not isinstance(binary_package, str):
            binary_package = binary_package.name

        package_transformation = self.package_transformations[binary_package]
        return self._source_condition_context.replace(
            binary_package=package_transformation.binary_package,
            substitution=package_transformation.substitution,
        )

    def apply_fs_transformations(
        self,
        package: str,
        fs_root: FSPath,
    ) -> None:
        if package in self._used_for:
            raise ValueError(
                f"data.tar contents for {package} has already been finalized!?"
            )
        if package not in self.package_transformations:
            raise ValueError(
                f'The package "{package}" was not relevant for the manifest!?'
            )
        package_transformation = self.package_transformations[package]
        condition_context = ConditionContext(
            binary_package=package_transformation.binary_package,
            substitution=package_transformation.substitution,
            deb_options_and_profiles=self._build_env,
            dpkg_architecture_variables=self._dpkg_architecture_variables,
            dpkg_arch_query_table=self._dpkg_arch_query_table,
        )
        norm_rules = list(
            builtin_mode_normalization_rules(
                self._dpkg_architecture_variables,
                package_transformation.binary_package,
                package_transformation.substitution,
            )
        )
        norm_mode_transformation_rule = ModeNormalizationTransformationRule(norm_rules)
        norm_mode_transformation_rule.transform_file_system(fs_root, condition_context)
        for transformation in package_transformation.transformations:
            transformation.run_transform_file_system(fs_root, condition_context)
        interpreter_normalization = NormalizeShebangLineTransformation()
        interpreter_normalization.transform_file_system(fs_root, condition_context)

    def finalize_data_tar_contents(
        self,
        package: str,
        fs_root: FSPath,
        clamp_mtime_to: int,
    ) -> IntermediateManifest:
        if package in self._used_for:
            raise ValueError(
                f"data.tar contents for {package} has already been finalized!?"
            )
        if package not in self.package_transformations:
            raise ValueError(
                f'The package "{package}" was not relevant for the manifest!?'
            )
        self._used_for.add(package)

        # At this point, there so be no further mutations to the file system (because the will not
        # be present in the intermediate manifest)
        cast("FSRootDir", fs_root).is_read_write = False

        intermediate_manifest = list(
            _generate_intermediate_manifest(
                fs_root,
                clamp_mtime_to,
            )
        )
        return intermediate_manifest

    def apply_to_binary_staging_directory(
        self,
        package: str,
        fs_root: FSPath,
        clamp_mtime_to: int,
    ) -> IntermediateManifest:
        self.apply_fs_transformations(package, fs_root)
        return self.finalize_data_tar_contents(package, fs_root, clamp_mtime_to)


@dataclasses.dataclass(slots=True)
class SearchDirOrderState:
    search_dir: VirtualPath
    applies_to: Union[Set[BinaryPackage], FrozenSet[BinaryPackage]] = dataclasses.field(
        default_factory=set
    )
    after: Set[str] = dataclasses.field(default_factory=set)


def _present_installation_dirs(
    search_dirs: Sequence[SearchDir],
    checked_missing_dirs: Sequence[VirtualPath],
    all_pkgs: FrozenSet[BinaryPackage],
) -> None:
    _info("The following directories are considered search dirs (in order):")
    max_len = max((len(s.search_dir.fs_path) for s in search_dirs), default=1)
    for search_dir in search_dirs:
        applies_to = ""
        if search_dir.applies_to < all_pkgs:
            names = ", ".join(p.name for p in search_dir.applies_to)
            applies_to = f"  [only applicable to: {names}]"
        remark = ""
        if not os.path.isdir(search_dir.search_dir.fs_path):
            remark = "  (skipped; absent)"
        _info(f" * {search_dir.search_dir.fs_path:{max_len}}{applies_to}{remark}")

    if checked_missing_dirs:
        _info('The following directories are considered for "not-installed" paths;')
        for d in checked_missing_dirs:
            remark = ""
            if not os.path.isdir(d.fs_path):
                remark = " (skipped; absent)"
            _info(f" * {d.fs_path:{max_len}}{remark}")


def _determine_search_dir_order(
    requested: Mapping[BinaryPackage, List[VirtualPath]],
    all_pkgs: FrozenSet[BinaryPackage],
    default_search_dirs: List[VirtualPath],
    source_root: VirtualPath,
) -> Sequence[SearchDir]:
    search_dir_table = {}
    assert requested.keys() <= all_pkgs
    for pkg in all_pkgs:
        paths = requested.get(pkg, default_search_dirs)
        previous_search_dir: Optional[SearchDirOrderState] = None
        for path in paths:
            try:
                search_dir_state = search_dir_table[path.fs_path]
            except KeyError:
                search_dir_state = SearchDirOrderState(path)
                search_dir_table[path.fs_path] = search_dir_state
            search_dir_state.applies_to.add(pkg)
            if previous_search_dir is not None:
                search_dir_state.after.add(previous_search_dir.search_dir.fs_path)
            previous_search_dir = search_dir_state

    search_dirs_in_order = []
    released = set()
    remaining = set()
    for search_dir_state in search_dir_table.values():
        if not (search_dir_state.after <= released):
            remaining.add(search_dir_state.search_dir.fs_path)
            continue
        search_dirs_in_order.append(search_dir_state)
        released.add(search_dir_state.search_dir.fs_path)

    while remaining:
        current_released = len(released)
        for fs_path in remaining:
            search_dir_state = search_dir_table[fs_path]
            if not search_dir_state.after.issubset(released):
                remaining.add(search_dir_state.search_dir.fs_path)
                continue
            search_dirs_in_order.append(search_dir_state)
            released.add(search_dir_state.search_dir.fs_path)

        if current_released == len(released):
            names = ", ".join(remaining)
            _error(
                f"There is a circular dependency (somewhere) between the search dirs: {names}."
                " Note that the search directories across all packages have to be ordered (and the"
                " source root should generally be last)"
            )
        remaining -= released

    search_dirs_in_order.append(
        SearchDirOrderState(
            source_root,
            all_pkgs,
        )
    )

    return tuple(
        # Avoid duplicating all_pkgs
        SearchDir(
            s.search_dir,
            frozenset(s.applies_to) if s.applies_to != all_pkgs else all_pkgs,
        )
        for s in search_dirs_in_order
    )



from .hexadecimal import Hex as Hex
from .library import Library as Library
from .utilities import Utilities as Utilities
from _typeshed import Incomplete
from typing import Any

TTY_AWARE: bool
IS_TTY: Incomplete

class Colored:
    def __init__(self, name: Any) -> None: ...
    def attribute(self, line_color: str = '') -> str: ...
    def foreground(self) -> str: ...
    def background(self) -> str: ...
    @staticmethod
    def enable_windows_terminal_mode() -> Any: ...
    @staticmethod
    def enabled() -> bool: ...

def style(name: int | str, color: str | int = '') -> str: ...
def fore(name: int | str) -> str: ...
def back(name: int | str) -> str: ...
def fore_rgb(r: int | str, g: int | str, b: int | str) -> str: ...
def back_rgb(r: int | str, g: int | str, b: int | str) -> str: ...
def attr(name: int | str) -> str: ...
def fg(name: int | str) -> str: ...
def bg(name: int | str) -> str: ...
def stylize(text: str, formatting: int | str, reset: bool = True) -> str: ...
def stylize_interactive(text: str, formatting: str, reset: bool = True) -> str: ...
def set_tty_aware(awareness: bool = True) -> None: ...



from typing import Mapping, Any, Optional, Union

import pytest
from debian.deb822 import Deb822

from debputy.plugin.api.test_api.test_impl import FAKE_DPKG_QUERY_TABLE
from debputy.yaml.compat import CommentedMap

from debputy.lsp.maint_prefs import (
    MaintainerPreferenceTable,
    determine_effective_preference,
    EffectiveFormattingPreference,
    _WAS_DEFAULTS,
)
from debputy.packages import SourcePackage


def test_load_styles() -> None:
    styles = MaintainerPreferenceTable.load_preferences()
    assert "niels@thykier.net" in styles.maintainer_preferences
    nt_maint_pref = styles.maintainer_preferences["niels@thykier.net"]
    # Note this is data dependent; if it fails because the style changes, update the test
    assert nt_maint_pref.canonical_name == "Niels Thykier"
    assert not nt_maint_pref.is_packaging_team
    black_style = styles.named_styles["black"]
    nt_style = nt_maint_pref.formatting
    assert nt_style is not None
    assert black_style == black_style


def test_load_no_styles() -> None:
    styles = MaintainerPreferenceTable.load_preferences()
    assert "packages@qa.debian.org" in styles.maintainer_preferences
    qa_maint_pref = styles.maintainer_preferences["packages@qa.debian.org"]
    assert qa_maint_pref.canonical_name == "Debian QA Group"
    assert qa_maint_pref.is_packaging_team
    # Orphaned packages do not have a predefined style, since Debian (nor Debian QA) have
    # one well-defined style.
    assert qa_maint_pref.formatting is None


def test_load_named_styles() -> None:
    styles = MaintainerPreferenceTable.load_preferences()
    assert "black" in styles.named_styles
    black_style = styles.named_styles["black"]
    # Note this is data dependent; if it fails because the style changes, update the test
    assert black_style.deb822_normalize_field_content
    assert black_style.deb822_short_indent
    assert black_style.deb822_always_wrap
    assert black_style.deb822_trailing_separator
    assert black_style.deb822_max_line_length == 79
    assert not black_style.deb822_normalize_stanza_order

    # TODO: Not implemented yet
    assert not black_style.deb822_normalize_field_order


def test_compat_styles() -> None:
    styles = MaintainerPreferenceTable.load_preferences()

    # Data dependent; if it breaks, provide a stubbed style preference table
    assert "niels@thykier.net" in styles.maintainer_preferences
    assert "zeha@debian.org" in styles.maintainer_preferences
    assert "random-package@packages.debian.org" not in styles.maintainer_preferences
    assert "random@example.org" not in styles.maintainer_preferences

    nt_style = styles.maintainer_preferences["niels@thykier.net"].formatting
    zeha_style = styles.maintainer_preferences["zeha@debian.org"].formatting

    # Data dependency
    assert nt_style == zeha_style

    fields = Deb822(
        {
            "Package": "foo",
            "Maintainer": "Foo <random-package@packages.debian.org>",
            "Uploaders": "Niels Thykier <niels@thykier.net>",
        },
    )
    src = SourcePackage(fields)

    effective_style, tool, _ = determine_effective_preference(styles, src, None)
    assert effective_style == nt_style
    assert tool == "debputy reformat"

    fields["Uploaders"] = (
        "Niels Thykier <niels@thykier.net>, Chris Hofstaedtler <zeha@debian.org>"
    )
    src = SourcePackage(fields)

    effective_style, tool, _ = determine_effective_preference(styles, src, None)
    assert effective_style == nt_style
    assert effective_style == zeha_style
    assert tool == "debputy reformat"

    fields["Uploaders"] = (
        "Niels Thykier <niels@thykier.net>, Chris Hofstaedtler <zeha@debian.org>, Random Developer <random@example.org>"
    )
    src = SourcePackage(fields)

    effective_style, tool, _ = determine_effective_preference(styles, src, None)
    assert effective_style is None
    assert tool is None


@pytest.mark.xfail
def test_compat_styles_team_maint() -> None:
    styles = MaintainerPreferenceTable.load_preferences()
    fields = Deb822(
        {
            "Package": "foo",
            # Missing a stubbed definition for `team@lists.debian.org`
            "Maintainer": "Packaging Team <team@lists.debian.org>",
            "Uploaders": "Random Developer <random@example.org>",
        },
    )
    src = SourcePackage(fields)
    assert "team@lists.debian.org" in styles.maintainer_preferences
    assert "random@example.org" not in styles.maintainer_preferences
    team_style = styles.maintainer_preferences["team@lists.debian.org"]
    assert team_style.is_packaging_team
    effective_style, tool, _ = determine_effective_preference(styles, src, None)
    assert effective_style == team_style.formatting
    assert tool is None


def test_x_style() -> None:
    styles = MaintainerPreferenceTable.load_preferences()
    fields = Deb822(
        {
            "Package": "foo",
            "X-Style": "black",
            "Maintainer": "Random Developer <random@example.org>",
        },
    )
    src = SourcePackage(fields)
    assert "random@example.org" not in styles.maintainer_preferences
    assert "black" in styles.named_styles
    black_style = styles.named_styles["black"]
    effective_style, tool, _ = determine_effective_preference(styles, src, None)
    assert effective_style == black_style
    assert tool == "debputy reformat"


@pytest.mark.parametrize(
    "variables",
    [
        {"SALSA_CI_DISABLE_WRAP_AND_SORT": "no"},
        {"SALSA_CI_DISABLE_WRAP_AND_SORT": 0},
        {"SALSA_CI_DISABLE_WRAP_AND_SORT": False},
        {"SALSA_CI_ENABLE_WRAP_AND_SORT": "yes"},
        {"SALSA_CI_ENABLE_WRAP_AND_SORT": 1},
        {"SALSA_CI_ENABLE_WRAP_AND_SORT": True},
    ],
)
def test_was_from_salsa_ci_style_enabled(variables: Mapping[str, Any]) -> None:
    styles = MaintainerPreferenceTable.load_preferences()
    fields = Deb822(
        {
            "Package": "foo",
            "Maintainer": "Random Developer <random@example.org>",
        },
    )
    src = SourcePackage(fields)
    assert "random@example.org" not in styles.maintainer_preferences
    effective_style, tool, _ = determine_effective_preference(styles, src, None)
    assert effective_style is None
    assert tool is None

    salsa_ci = CommentedMap({"variables": CommentedMap(variables)})
    effective_style, tool, _ = determine_effective_preference(styles, src, salsa_ci)
    was_style = EffectiveFormattingPreference(**_WAS_DEFAULTS)
    assert effective_style == was_style
    assert tool == "wrap-and-sort"


@pytest.mark.parametrize(
    "variables",
    [
        {},
        {"SALSA_CI_DISABLE_WRAP_AND_SORT": "yes"},
        {"SALSA_CI_DISABLE_WRAP_AND_SORT": 1},
        {"SALSA_CI_DISABLE_WRAP_AND_SORT": True},
        {"SALSA_CI_DISABLE_WRAP_AND_SORT": {}},
        {"SALSA_CI_DISABLE_WRAP_AND_SORT": []},
        {"SALSA_CI_DISABLE_WRAP_AND_SORT": None},
        {"SALSA_CI_ENABLE_WRAP_AND_SORT": "no"},
        {"SALSA_CI_ENABLE_WRAP_AND_SORT": 0},
        {"SALSA_CI_ENABLE_WRAP_AND_SORT": False},
        {"SALSA_CI_ENABLE_WRAP_AND_SORT": {}},
        {"SALSA_CI_ENABLE_WRAP_AND_SORT": []},
        {"SALSA_CI_ENABLE_WRAP_AND_SORT": None},
        {
            "SALSA_CI_DISABLE_WRAP_AND_SORT": "yes",
            "SALSA_CI_ENABLE_WRAP_AND_SORT": "yes",
        },
        {"SALSA_CI_DISABLE_WRAP_AND_SORT": 1, "SALSA_CI_ENABLE_WRAP_AND_SORT": 1},
        {"SALSA_CI_DISABLE_WRAP_AND_SORT": True, "SALSA_CI_ENABLE_WRAP_AND_SORT": True},
        {"SALSA_CI_DISABLE_WRAP_AND_SORT": {}, "SALSA_CI_ENABLE_WRAP_AND_SORT": True},
        {"SALSA_CI_DISABLE_WRAP_AND_SORT": [], "SALSA_CI_ENABLE_WRAP_AND_SORT": True},
        {"SALSA_CI_DISABLE_WRAP_AND_SORT": None, "SALSA_CI_ENABLE_WRAP_AND_SORT": True},
    ],
)
def test_was_from_salsa_ci_style_disabled(variables: Mapping[str, Any]) -> None:
    styles = MaintainerPreferenceTable.load_preferences()
    fields = Deb822(
        {
            "Package": "foo",
            "Maintainer": "Random Developer <random@example.org>",
        },
    )
    src = SourcePackage(fields)
    assert "random@example.org" not in styles.maintainer_preferences
    effective_style, tool, _ = determine_effective_preference(styles, src, None)
    assert effective_style is None
    assert tool is None
    salsa_ci = CommentedMap({"variables": CommentedMap(variables)})
    effective_style, tool, _ = determine_effective_preference(styles, src, salsa_ci)
    assert effective_style is None
    assert tool is None


@pytest.mark.parametrize(
    "was_args,style_delta",
    [
        (
            "-a",
            {
                "deb822_always_wrap": True,
            },
        ),
        (
            "-sa",
            {
                "deb822_always_wrap": True,
                "deb822_short_indent": True,
            },
        ),
        (
            "-sa --keep-first",
            {
                "deb822_always_wrap": True,
                "deb822_short_indent": True,
            },
        ),
        (
            "-sab --keep-first",
            {
                "deb822_always_wrap": True,
                "deb822_short_indent": True,
                "deb822_normalize_stanza_order": True,
            },
        ),
        (
            "-sab --no-keep-first",
            {
                "deb822_always_wrap": True,
                "deb822_short_indent": True,
                "deb822_normalize_stanza_order": False,
            },
        ),
    ],
)
def test_was_from_salsa_ci_style_args(
    was_args: str,
    style_delta: Optional[Mapping[str, Any]],
) -> None:
    styles = MaintainerPreferenceTable.load_preferences()
    fields = Deb822(
        {
            "Package": "foo",
            "Maintainer": "Random Developer <random@example.org>",
        },
    )
    src = SourcePackage(fields)
    assert "random@example.org" not in styles.maintainer_preferences
    salsa_ci = CommentedMap(
        {
            "variables": CommentedMap(
                {
                    "SALSA_CI_DISABLE_WRAP_AND_SORT": "no",
                    "SALSA_CI_WRAP_AND_SORT_ARGS": was_args,
                }
            )
        }
    )
    effective_style, tool, _ = determine_effective_preference(styles, src, salsa_ci)
    if style_delta is None:
        assert effective_style is None
        assert tool is None
    else:
        was_style = EffectiveFormattingPreference(**_WAS_DEFAULTS).replace(
            **style_delta,
        )

        assert effective_style == was_style
        assert tool == f"wrap-and-sort {was_args}".strip()



from .colored import Colored as Colored

def cprint(text: str, fore_256: int | str = '', back_256: int | str = '', fore_rgb: tuple = (255, 255, 255), back_rgb: tuple = (0, 0, 0), formatting: int | str = '', line_color: int | str = '', reset: bool = True, **kwargs) -> None: ...



import textwrap

import pytest

from debputy.lsp.lsp_debian_copyright import _lint_debian_copyright
from debputy.packages import DctrlParser
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from debputy.plugin.api.test_api import build_virtual_file_system
from lint_tests.lint_tutil import (
    group_diagnostics_by_severity,
    LintWrapper,
)

from debputy.lsprotocol.types import DiagnosticSeverity


@pytest.fixture
def line_linter(
    debputy_plugin_feature_set: PluginProvidedFeatureSet,
    lint_dctrl_parser: DctrlParser,
) -> LintWrapper:
    return LintWrapper(
        "/nowhere/debian/copyright",
        _lint_debian_copyright,
        debputy_plugin_feature_set,
        lint_dctrl_parser,
    )


def test_dcpy_files_lint(line_linter: LintWrapper) -> None:
    lines = textwrap.dedent(
        """\
    Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/

    Files: foo .//unnecessary///many/slashes
    Copyright: Noone <noone@example.com>
    License: something
     yada yada yada
    """
    ).splitlines(keepends=True)

    diagnostics = line_linter(lines)
    by_severity = group_diagnostics_by_severity(diagnostics)
    assert DiagnosticSeverity.Warning in by_severity

    assert DiagnosticSeverity.Error not in by_severity
    assert DiagnosticSeverity.Hint not in by_severity
    assert DiagnosticSeverity.Information not in by_severity

    warnings = by_severity[DiagnosticSeverity.Warning]
    print(warnings)
    assert len(warnings) == 2

    first_warn, second_warn = warnings

    msg = 'Unnecessary prefix ".//"'
    assert first_warn.message == msg
    assert f"{first_warn.range}" == "2:11-2:14"

    msg = 'Simplify to a single "/"'
    assert second_warn.message == msg
    assert f"{second_warn.range}" == "2:25-2:28"


def test_dcpy_files_matches_dir_lint(line_linter: LintWrapper) -> None:
    lines = textwrap.dedent(
        """\
    Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/

    Files: foo
    Copyright: Noone <noone@example.com>
    License: something
     yada yada yada
    """
    ).splitlines(keepends=True)

    source_root = build_virtual_file_system(["./foo/bar"])
    line_linter.source_root = source_root

    diagnostics = line_linter(lines)
    assert len(diagnostics) == 1
    issue = diagnostics[0]

    msg = "Directories cannot be a match. Use `dir/*` to match everything in it"
    assert issue.message == msg
    assert f"{issue.range}" == "2:7-2:10"
    assert issue.severity == DiagnosticSeverity.Warning


def test_dcpy_anon_real_example_lint(line_linter: LintWrapper) -> None:
    lines = textwrap.dedent(
        """\
    Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
    Source: https://github.com/some-where/some-project
    Upstream-Name: some-project
    Upstream-Contact: https://github.com/some-where/some-project/issues

    Files: *
    Copyright:
     2022-2024 Some One <some-one@example.com>
     2022-2024 some-project contributors
    License: Expat

    Files: debian/*
    Copyright: 2024 Debian Contributor <debian.contributor@example.com>
    License: Expat
    Comment: Debian packaging is licensed under the same terms as upstream

    License: Expat
     Permission is hereby granted, free of charge, to any person obtaining
     a copy of this software and associated documentation files (the
     "Software"), to deal in the Software without restriction, including
     without limitation the rights to use, copy, modify, merge, publish,
     distribute, sublicense, and/or sell copies of the Software, and to
     permit persons to whom the Software is furnished to do so, subject to
     the following conditions:
     .
     The above copyright notice and this permission notice shall be included
     in all copies or substantial portions of the Software.
     .
     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
     EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
     MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
     IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
     CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
     TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
     SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    """
    ).splitlines(keepends=True)

    diagnostics = line_linter(lines)
    assert not diagnostics


def test_not_dcpy_syntax_error_before_format(line_linter: LintWrapper) -> None:
    lines = textwrap.dedent(
        """\
    Random error

    Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
    Source: https://github.com/some-where/some-project
    Upstream-Name: some-project
    Upstream-Contact: https://github.com/some-where/some-project/issues
"""
    ).splitlines(keepends=True)

    diagnostics = line_linter(lines)
    assert not diagnostics


def test_not_dcpy_no_format_field(line_linter: LintWrapper) -> None:
    lines = textwrap.dedent(
        """\
    Some free-text debian/copyright file
"""
    ).splitlines(keepends=True)

    diagnostics = line_linter(lines)
    assert not diagnostics



import textwrap

from debputy.installations import InstallRule
from debputy.maintscript_snippet import DpkgMaintscriptHelperCommand
from debputy.manifest_conditions import ManifestCondition
from debputy.plugin.api import reference_documentation
from debputy.plugin.debputy.to_be_api_types import BuildRule
from debputy.transformation_rules import TransformationRule
from debputy.util import manifest_format_doc

SUPPORTED_DISPATCHABLE_TABLE_PARSERS = {
    InstallRule: "installations",
    TransformationRule: "packages.{{PACKAGE}}.transformations",
    DpkgMaintscriptHelperCommand: "packages.{{PACKAGE}}.conffile-management",
    ManifestCondition: "*.when",
    BuildRule: "builds",
}

OPARSER_MANIFEST_ROOT = "<ROOT>"
OPARSER_PACKAGES_ROOT = "packages"
OPARSER_PACKAGES = "packages.{{PACKAGE}}"
OPARSER_MANIFEST_DEFINITIONS = "definitions"

SUPPORTED_DISPATCHABLE_OBJECT_PARSERS = {
    OPARSER_MANIFEST_ROOT: reference_documentation(
        reference_documentation_url=manifest_format_doc(""),
    ),
    OPARSER_MANIFEST_DEFINITIONS: reference_documentation(
        title="Packager provided definitions",
        description="Reusable packager provided definitions such as manifest variables.",
        reference_documentation_url=manifest_format_doc(
            "packager-provided-definitions"
        ),
    ),
    OPARSER_PACKAGES: reference_documentation(
        title="Binary package rules",
        description=textwrap.dedent(
            """\
            Inside the manifest, the `packages` mapping can be used to define requests for the binary packages
            you want `debputy` to produce.  Each key inside `packages` must be the name of a binary package
            defined in `debian/control`.  The value is a dictionary defining which features that `debputy`
            should apply to that binary package.  An example could be:

                packages:
                    foo:
                        transformations:
                            - create-symlink:
                                  path: usr/share/foo/my-first-symlink
                                  target: /usr/share/bar/symlink-target
                            - create-symlink:
                                  path: usr/lib/{{DEB_HOST_MULTIARCH}}/my-second-symlink
                                  target: /usr/lib/{{DEB_HOST_MULTIARCH}}/baz/symlink-target
                    bar:
                        transformations:
                        - create-directories:
                           - some/empty/directory.d
                           - another/empty/integration-point.d
                        - create-directories:
                             path: a/third-empty/directory.d
                             owner: www-data
                             group: www-data

            In this case, `debputy` will create some symlinks inside the `foo` package and some directories for
            the `bar` package.  The following subsections define the keys you can use under each binary package.
        """
        ),
        reference_documentation_url=manifest_format_doc("binary-package-rules"),
    ),
}



import json
import os.path
from typing import List, Tuple, Type, cast, TYPE_CHECKING

import pytest

from debputy.exceptions import DebputyFSIsROError
from debputy.plugin.api import (
    DebputyPluginInitializer,
    BinaryCtrlAccessor,
    PackageProcessingContext,
    VirtualPath,
    virtual_path_def,
)
from debputy.exceptions import PluginConflictError, PluginAPIViolationError
from debputy.plugin.api.impl import DebputyPluginInitializerProvider
from debputy.plugin.api.impl_types import automatic_discard_rule_example
from debputy.plugin.api.test_api import (
    build_virtual_file_system,
    package_metadata_context,
    initialize_plugin_under_test,
)
from debputy.plugin.api.test_api.test_impl import (
    initialize_plugin_under_test_preloaded,
    initialize_plugin_under_test_from_inline_json,
)

if TYPE_CHECKING:
    from debputy.plugin.api import PluginInitializationEntryPoint

    # Irrelevant, but makes the import not "unused" for things that does not parse `cast("...", ...)` expressions
    assert PluginInitializationEntryPoint is not None


CUSTOM_PLUGIN_JSON_FILE = os.path.join(
    os.path.dirname(__file__), "data", "custom-plugin.json.in"
)


def bad_metadata_detector_fs_rw(
    path: VirtualPath,
    _ctrl: BinaryCtrlAccessor,
    _context: PackageProcessingContext,
) -> None:
    del path["foo"]


def conflicting_plugin(api: DebputyPluginInitializer) -> None:
    api.packager_provided_file(
        "logrotate",
        "/some/where/that/is/not/etc/logrotate.d/{name}",
    )


def bad_plugin(api: DebputyPluginInitializer) -> None:
    api.metadata_or_maintscript_detector("fs_rw", bad_metadata_detector_fs_rw)


def adr_inconsistent_example_plugin(api: DebputyPluginInitializerProvider) -> None:
    api.automatic_discard_rule(
        "adr-example-test",
        lambda p: p.name == "discard-me",
        examples=automatic_discard_rule_example(
            "foo/discard-me",
            ("bar/discard-me", False),
            "baz/something",
            ("discard-me/foo", False),
        ),
    )


def test_conflict_with_debputy():
    with pytest.raises(PluginConflictError) as e_info:
        initialize_plugin_under_test_preloaded(
            1,
            conflicting_plugin,
            plugin_name="conflicting-plugin",
        )
    message = (
        'The stem "logrotate" is registered twice for packager provided files.'
        " Once by debputy and once by conflicting-plugin"
    )
    assert message == e_info.value.args[0]


def test_metadata_read_only():
    plugin = initialize_plugin_under_test_preloaded(
        1,
        bad_plugin,
        plugin_name="bad-plugin",
    )
    fs = build_virtual_file_system(["./foo"])
    with pytest.raises(PluginAPIViolationError) as e_info:
        plugin.run_metadata_detector("fs_rw", fs)

    assert isinstance(e_info.value.__cause__, DebputyFSIsROError)


def test_packager_provided_files():
    plugin = initialize_plugin_under_test(plugin_desc_file=CUSTOM_PLUGIN_JSON_FILE)
    assert plugin.packager_provided_files_by_stem().keys() == {
        "my-file",
        "test-file-from-json",
    }
    my_file = [p for p in plugin.packager_provided_files() if p.stem == "my-file"][0]

    assert my_file.stem == "my-file"
    assert my_file.compute_dest("g++-3.1")[1] == "g__-3.1.conf"


def test_path_checks():
    symlink_path = "./foo"
    with pytest.raises(ValueError) as e_info:
        virtual_path_def(symlink_path, link_target="/bar", mode=0o0755)
    assert (
        e_info.value.args[0]
        == f'Please do not provide mode for symlinks. Triggered by "{symlink_path}"'
    )
    with pytest.raises(ValueError) as e_info:
        virtual_path_def(symlink_path + "/", link_target="/bar")
    msg = (
        "Path name looks like a directory, but a symlink target was also provided."
        f' Please remove the trailing slash OR the symlink_target. Triggered by "{symlink_path}/"'
    )
    assert e_info.value.args[0] == msg


def test_metadata_detector_applies_to_check():
    plugin_name = "custom-plugin"
    metadata_detector_id = "udeb-only"
    plugin = initialize_plugin_under_test(plugin_desc_file=CUSTOM_PLUGIN_JSON_FILE)
    with pytest.raises(ValueError) as e_info:
        plugin.run_metadata_detector(
            metadata_detector_id,
            build_virtual_file_system(["./usr/share/doc/foo/copyright"]),
        )
    msg = f'The detector "{metadata_detector_id}" from {plugin_name} does not apply to the given package.'
    assert e_info.value.args[0].startswith(msg)

    metadata = plugin.run_metadata_detector(
        metadata_detector_id,
        build_virtual_file_system(["./usr/share/doc/foo/copyright"]),
        context=package_metadata_context(package_fields={"Package-Type": "udeb"}),
    )
    assert metadata.substvars["Test:Udeb-Metadata-Detector"] == "was-run"


@pytest.mark.parametrize(
    "variables,exec_type",
    [
        (
            [("DEBPUTY_VAR", "RESERVED")],
            ValueError,
        ),
        (
            [("_DEBPUTY_VAR", "RESERVED")],
            ValueError,
        ),
        (
            [("_FOO", "RESERVED")],
            ValueError,
        ),
        (
            [("path:_var", "RESERVED")],
            ValueError,
        ),
        (
            [("path:DEBPUTY_VAR", "RESERVED")],
            ValueError,
        ),
        (
            [("DEB_VAR", "RESERVED")],
            ValueError,
        ),
        (
            [("DPKG_VAR", "RESERVED")],
            ValueError,
        ),
        (
            [("PACKAGE", "RESERVED")],
            ValueError,
        ),
        (
            [("foo:var", "RESERVED")],
            ValueError,
        ),
        (
            [("env:var", "RESERVED")],
            ValueError,
        ),
        (
            [("SOURCE_DATE_EPOCH", "RESERVED")],
            ValueError,
        ),
        (
            [("!MY_VAR", "INVALID_NAME")],
            ValueError,
        ),
        (
            [("VALUE_DEPENDS_ON_VAR", "{{UNKNOWN_VAR}}")],
            ValueError,
        ),
        (
            [("VALUE_DEPENDS_ON_VAR", "{{DEB_HOST_ARCH}}")],
            ValueError,
        ),
        (
            [("DEFINED_TWICE", "ONCE"), ("DEFINED_TWICE", "TWICE")],
            PluginConflictError,
        ),
    ],
)
def test_invalid_manifest_variables(
    variables: List[Tuple[str, str]],
    exec_type: Type[Exception],
) -> None:
    def _initializer(api: DebputyPluginInitializer):
        with pytest.raises(exec_type):
            for varname, value in variables:
                api.manifest_variable(varname, value)

    initialize_plugin_under_test_preloaded(
        1,
        _initializer,
        plugin_name="test-plugin",
        load_debputy_plugin=False,
    )


def test_valid_manifest_variables() -> None:
    variables = {
        "PLUGIN_VAR": "TEST VALUE",
        "ANOTHER_PLUGIN_VAR": "ANOTHER VALUE",
        "path:SOMEWHERE_DIR": "/usr/share/some-where",
    }

    def _initializer(api: DebputyPluginInitializer):
        for k, v in variables.items():
            api.manifest_variable(k, v)

    plugin = initialize_plugin_under_test_preloaded(
        1,
        _initializer,
        plugin_name="test-plugin",
        load_debputy_plugin=False,
    )

    assert plugin.declared_manifest_variables == variables.keys()


def test_valid_manifest_variables_json() -> None:
    variables = {
        "PLUGIN_VAR": "TEST VALUE",
        "ANOTHER_PLUGIN_VAR": "ANOTHER VALUE",
        "path:SOMEWHERE_DIR": "/usr/share/some-where",
    }
    content = {
        "api-compat-version": 1,
        "manifest-variables": [
            {
                "name": k,
                "value": v,
            }
            for k, v in variables.items()
        ],
    }
    plugin = initialize_plugin_under_test_from_inline_json(
        "test-plugin",
        json.dumps(content),
    )
    assert plugin.declared_manifest_variables == variables.keys()


def test_automatic_discard_rules_example() -> None:
    plugin = initialize_plugin_under_test_preloaded(
        1,
        # Internal API used
        cast("PluginInitializationEntryPoint", adr_inconsistent_example_plugin),
        # API is restricted
        plugin_name="debputy",
        load_debputy_plugin=False,
    )
    issues = plugin.automatic_discard_rules_examples_with_issues()
    assert len(issues) == 1
    issue = issues[0]
    assert issue.name == "adr-example-test"
    assert issue.example_index == 0
    assert set(issue.inconsistent_paths) == {
        "/discard-me/foo",
        "/bar/discard-me",
        "/baz/something",
    }



import re
from typing import (
    Union,
    Sequence,
    Optional,
    List,
)

from debputy.linting.lint_util import LintState
from debputy.lsp.debputy_ls import DebputyLanguageServer
from debputy.lsp.lsp_debian_control_reference_data import (
    Dep5FileMetadata,
)
from debputy.lsp.lsp_features import (
    lint_diagnostics,
    lsp_completer,
    lsp_hover,
    lsp_standard_handler,
    lsp_folding_ranges,
    lsp_semantic_tokens_full,
    lsp_will_save_wait_until,
    lsp_format_document,
    SecondaryLanguage,
    LanguageDispatchRule,
)
from debputy.lsp.lsp_generic_deb822 import (
    deb822_completer,
    deb822_hover,
    deb822_folding_ranges,
    deb822_semantic_tokens_full,
    deb822_format_file,
    scan_for_syntax_errors_and_token_level_diagnostics,
)
from debputy.lsp.vendoring._deb822_repro import (
    Deb822FileElement,
    Deb822ParagraphElement,
)
from debputy.lsprotocol.types import (
    CompletionItem,
    CompletionList,
    CompletionParams,
    HoverParams,
    Hover,
    TEXT_DOCUMENT_CODE_ACTION,
    SemanticTokens,
    SemanticTokensParams,
    FoldingRangeParams,
    FoldingRange,
    WillSaveTextDocumentParams,
    TextEdit,
    DocumentFormattingParams,
)

try:
    from debputy.lsp.vendoring._deb822_repro.locatable import (
        Position as TEPosition,
        Range as TERange,
    )

    from pygls.server import LanguageServer
    from pygls.workspace import TextDocument
except ImportError:
    pass


_CONTAINS_SPACE_OR_COLON = re.compile(r"[\s:]")

_DISPATCH_RULE = LanguageDispatchRule.new_rule(
    "debian/copyright",
    "debian/copyright",
    [
        # emacs's name
        SecondaryLanguage("debian-copyright"),
        # vim's name
        SecondaryLanguage("debcopyright"),
    ],
)

_DEP5_FILE_METADATA = Dep5FileMetadata()

lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_CODE_ACTION)


@lsp_hover(_DISPATCH_RULE)
def _debian_copyright_hover(
    ls: "DebputyLanguageServer",
    params: HoverParams,
) -> Optional[Hover]:
    return deb822_hover(ls, params, _DEP5_FILE_METADATA)


@lsp_completer(_DISPATCH_RULE)
def _debian_copyright_completions(
    ls: "DebputyLanguageServer",
    params: CompletionParams,
) -> Optional[Union[CompletionList, Sequence[CompletionItem]]]:
    return deb822_completer(ls, params, _DEP5_FILE_METADATA)


@lsp_folding_ranges(_DISPATCH_RULE)
def _debian_copyright_folding_ranges(
    ls: "DebputyLanguageServer",
    params: FoldingRangeParams,
) -> Optional[Sequence[FoldingRange]]:
    return deb822_folding_ranges(ls, params, _DEP5_FILE_METADATA)


def _looks_like_a_dep5_file(
    deb822_file: Deb822FileElement,
    stanzas: List[Deb822ParagraphElement],
) -> bool:
    if not stanzas or "Format" not in stanzas[0]:
        # No parseable stanzas or the first one did not have a Format, which is necessary.
        return False

    for part in deb822_file.iter_parts():
        if part.is_error:
            # Error first, then it might just be a "Format:" in the middle of a free-text file.
            return False
        if isinstance(part, Deb822ParagraphElement):
            break
    return True


@lint_diagnostics(_DISPATCH_RULE)
async def _lint_debian_copyright(lint_state: LintState) -> None:
    deb822_file = lint_state.parsed_deb822_file_content
    stanzas = list(deb822_file)

    if not _looks_like_a_dep5_file(deb822_file, stanzas):
        return

    first_error = await scan_for_syntax_errors_and_token_level_diagnostics(
        deb822_file,
        lint_state,
    )
    header_stanza, files_stanza, _ = _DEP5_FILE_METADATA.stanza_types()

    async for paragraph_no, paragraph in lint_state.slow_iter(
        enumerate(stanzas, start=1)
    ):
        paragraph_pos = paragraph.position_in_file()
        if paragraph_pos.line_position >= first_error:
            break
        is_files_or_license_stanza = paragraph_no != 1
        if is_files_or_license_stanza:
            stanza_metadata = _DEP5_FILE_METADATA.classify_stanza(
                paragraph,
                paragraph_no,
            )
            other_stanza_metadata = header_stanza
            other_stanza_name = "Header"
        elif "Format" in paragraph:
            stanza_metadata = header_stanza
            other_stanza_metadata = files_stanza
            other_stanza_name = "Files/License"
        else:
            break

        await stanza_metadata.stanza_diagnostics(
            deb822_file,
            paragraph,
            paragraph_pos,
            lint_state,
            confusable_with_stanza_name=other_stanza_name,
            confusable_with_stanza_metadata=other_stanza_metadata,
        )


@lsp_will_save_wait_until(_DISPATCH_RULE)
def _debian_copyright_on_save_formatting(
    ls: "DebputyLanguageServer",
    params: WillSaveTextDocumentParams,
) -> Optional[Sequence[TextEdit]]:
    doc = ls.workspace.get_text_document(params.text_document.uri)
    lint_state = ls.lint_state(doc)
    return deb822_format_file(lint_state, _DEP5_FILE_METADATA)


def _reformat_debian_copyright(
    lint_state: LintState,
) -> Optional[Sequence[TextEdit]]:
    return deb822_format_file(lint_state, _DEP5_FILE_METADATA)


@lsp_format_document(_DISPATCH_RULE)
def _debian_copyright_on_save_formatting(
    ls: "DebputyLanguageServer",
    params: DocumentFormattingParams,
) -> Optional[Sequence[TextEdit]]:
    doc = ls.workspace.get_text_document(params.text_document.uri)
    lint_state = ls.lint_state(doc)
    return deb822_format_file(lint_state, _DEP5_FILE_METADATA)


@lsp_semantic_tokens_full(_DISPATCH_RULE)
async def _debian_copyright_semantic_tokens_full(
    ls: "DebputyLanguageServer",
    request: SemanticTokensParams,
) -> Optional[SemanticTokens]:
    return await deb822_semantic_tokens_full(
        ls,
        request,
        _DEP5_FILE_METADATA,
    )
